# import asyncio
# import concurrent.futures
# import threading
# import functools
# import inspect
# from typing import Optional, Dict, Any, Union
# from geopandas import GeoDataFrame
# from shapely.geometry.base import BaseGeometry as ShapelyGeometry
# from .async_client import AsyncClient
# from typing import TYPE_CHECKING

# # # Add this after your other imports
# # if TYPE_CHECKING:
# #     from .endpoints.dataset_management import DatasetManagement
# #     from .endpoints.user_management import UserManagement
# #     from .endpoints.mass_stats import MassStats
# #     from .endpoints.group_management import GroupManagement
# #     from .endpoints.space_management import SpaceManagement
# #     from .endpoints.model_management import ModelManagement
# #     from .endpoints.auth import AuthClient


# class SyncWrapper:
#     """Generic synchronous wrapper with __dir__ support for runtime autocomplete."""
    
#     def __init__(self, async_obj, sync_client):
#         self._async_obj = async_obj
#         self._sync_client = sync_client
    
#     def __dir__(self):
#         """Return list of attributes for autocomplete in interactive environments."""
#         async_attrs = [attr for attr in dir(self._async_obj) if not attr.startswith('_')]
#         wrapper_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
#         return list(set(async_attrs + wrapper_attrs))
    
#     def __getattr__(self, name):
#         """Dynamically wrap any method call to convert async to sync."""
#         attr = getattr(self._async_obj, name)
        
#         if callable(attr):
#             @functools.wraps(attr)
#             def sync_wrapper(*args, **kwargs):
#                 result = attr(*args, **kwargs)
#                 if hasattr(result, '__await__'):
#                     return self._sync_client._run_async(result)
#                 return result
#             return sync_wrapper
        
#         return attr


# class SyncClient:
#     """
#     Thread-safe synchronous wrapper for AsyncClient.
#     Uses a persistent event loop in a dedicated thread to avoid event loop conflicts.
#     """

#     # datasets: 'DatasetManagement'
#     # users: 'UserManagement' 
#     # mass_stats: 'MassStats'
#     # groups: 'GroupManagement'
#     # space: 'SpaceManagement'
#     # model: 'ModelManagement'
#     # auth: 'AuthClient'
    
#     def __init__(self, url: Optional[str] = None, api_key: Optional[str] = None, verbose: bool = False):
#         self._async_client = AsyncClient(url=url, api_key=api_key, verbose=verbose)
#         self._context_entered = False
#         self._closed = False
        
#         # Thread and event loop management
#         self._loop = None
#         self._thread = None
#         self._loop_ready = None
#         self._loop_exception = None
        
#         # Initialize endpoint managers
#         print("we are here!!!!!!!!!!!!!!!!!")
#         self.datasets = SyncWrapper(self._async_client.datasets, self)
#         self.users = SyncWrapper(self._async_client.users, self)
#         self.mass_stats = SyncWrapper(self._async_client.mass_stats, self)
#         self.groups = SyncWrapper(self._async_client.groups, self)
#         self.space = SyncWrapper(self._async_client.space, self)
#         self.model = SyncWrapper(self._async_client.model, self)
#         self.auth = SyncWrapper(self._async_client.auth, self)
        
#         # Register cleanup
#         import atexit
#         atexit.register(self._cleanup)
    
#     def _ensure_event_loop(self):
#         """Ensure we have a persistent event loop in a dedicated thread."""
#         if self._loop is None or self._loop.is_closed():
#             self._loop_ready = threading.Event()
#             self._loop_exception = None
            
#             def run_loop():
#                 """Run the event loop in a dedicated thread."""
#                 try:
#                     # Create a new event loop for this thread
#                     self._loop = asyncio.new_event_loop()
#                     asyncio.set_event_loop(self._loop)
                    
#                     # Signal that the loop is ready
#                     self._loop_ready.set()
                    
#                     # Run the loop forever (until stopped)
#                     self._loop.run_forever()
#                 except Exception as e:
#                     self._loop_exception = e
#                     self._loop_ready.set()
#                 finally:
#                     # Clean up when the loop stops
#                     if self._loop and not self._loop.is_closed():
#                         self._loop.close()
            
#             # Start the thread
#             self._thread = threading.Thread(target=run_loop, daemon=True)
#             self._thread.start()
            
#             # Wait for the loop to be ready
#             self._loop_ready.wait(timeout=10)
            
#             if self._loop_exception:
#                 raise self._loop_exception
            
#             if not self._loop_ready.is_set():
#                 raise RuntimeError("Event loop failed to start within timeout")
    
#     def _run_async(self, coro):
#         """
#         Run async coroutine using persistent event loop.
#         This is the core method that makes everything work.
#         """
#         # Ensure we have an event loop
#         self._ensure_event_loop()
        
#         if self._loop.is_closed():
#             raise RuntimeError("Event loop is closed")
        
#         # Create a future to get the result back from the event loop thread
#         future = concurrent.futures.Future()
        
#         async def run_with_context():
#             """Run the coroutine with proper context management."""
#             try:
#                 # Ensure the async client is properly initialized
#                 await self._ensure_context()
                
#                 # Run the actual coroutine
#                 result = await coro
                
#                 # Set the result on the future
#                 future.set_result(result)
#             except Exception as e:
#                 # Set the exception on the future
#                 future.set_exception(e)
        
#         # Schedule the coroutine on the persistent event loop
#         self._loop.call_soon_threadsafe(
#             lambda: asyncio.create_task(run_with_context())
#         )
        
#         # Wait for the result (with timeout to avoid hanging)
#         try:
#             return future.result(timeout=300)  # 5 minute timeout
#         except concurrent.futures.TimeoutError:
#             raise RuntimeError("Async operation timed out after 5 minutes")
    
#     async def _ensure_context(self):
#         """Ensure the async client context is entered."""
#         if not self._context_entered and not self._closed:
#             await self._async_client.__aenter__()
#             self._context_entered = True
    
#     async def _exit_context(self):
#         """Exit the async client context."""
#         if self._context_entered and not self._closed:
#             await self._async_client.__aexit__(None, None, None)
#             self._context_entered = False
    
#     def close(self):
#         """Close the underlying async client session and stop the event loop."""
#         if not self._closed:
#             if self._loop and not self._loop.is_closed():
#                 # Schedule cleanup on the event loop
#                 future = concurrent.futures.Future()
                
#                 async def cleanup():
#                     """Clean up the async client."""
#                     try:
#                         await self._exit_context()
#                         future.set_result(None)
#                     except Exception as e:
#                         future.set_exception(e)
                
#                 # Run cleanup
#                 self._loop.call_soon_threadsafe(
#                     lambda: asyncio.create_task(cleanup())
#                 )
                
#                 # Wait for cleanup to complete
#                 try:
#                     future.result(timeout=10)
#                 except:
#                     pass  # Ignore cleanup errors
                
#                 # Stop the event loop
#                 self._loop.call_soon_threadsafe(self._loop.stop)
                
#                 # Wait for thread to finish
#                 if self._thread and self._thread.is_alive():
#                     self._thread.join(timeout=5)
            
#             self._closed = True
    
#     def _cleanup(self):
#         """Internal cleanup method called by atexit."""
#         if not self._closed:
#             try:
#                 self.close()
#             except Exception:
#                 pass  # Ignore cleanup errors
    
#     def __dir__(self):
#         """Return list of attributes for autocomplete in interactive environments."""
#         default_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
#         async_client_attrs = [attr for attr in dir(self._async_client) if not attr.startswith('_')]
#         endpoint_attrs = ['datasets', 'users', 'mass_stats', 'groups', 'space', 'model', 'auth']
#         all_attrs = default_attrs + async_client_attrs + endpoint_attrs
#         return list(set(all_attrs))
    
#     # Your existing methods (geoquery, zonal_stats, etc.)
#     def geoquery(
#         self,
#         expr: str,
#         feature: Union[Dict[str, Any], ShapelyGeometry],
#         in_crs: str = "epsg:4326",
#         out_crs: str = "epsg:4326",
#         resolution: int = -1,
#         geom_fix: bool = False,
#         **kwargs
#     ):
#         """Compute WCS query for a single geometry (synchronous version)."""
#         coro = self._async_client.geoquery(
#             expr=expr,
#             feature=feature,
#             in_crs=in_crs,
#             out_crs=out_crs,
#             output="netcdf",
#             resolution=resolution,
#             geom_fix=geom_fix,
#             **kwargs
#         )
#         return self._run_async(coro)

#     def zonal_stats(
#         self,
#         gdf: GeoDataFrame,
#         expr: str,
#         conc: int = 20,
#         inplace: bool = False,
#         in_crs: str = "epsg:4326",
#         out_crs: str = "epsg:4326",
#         resolution: int = -1,
#         geom_fix: bool = False,
#         drop_nan: bool = False,
#         spatial_reduction: str = None,
#         temporal_reduction: str = None,
#         max_memory_mb: int = 500,
#         stream_to_disk: bool = False,
#     ):
#         """Compute zonal statistics for all geometries in a GeoDataFrame (synchronous version)."""
#         coro = self._async_client.zonal_stats(
#             gdf=gdf,
#             expr=expr,
#             conc=conc,
#             inplace=inplace,
#             in_crs=in_crs,
#             out_crs=out_crs,
#             resolution=resolution,
#             geom_fix=geom_fix,
#             drop_nan=drop_nan,
#             spatial_reduction=spatial_reduction,
#             temporal_reduction=temporal_reduction,
#             max_memory_mb=max_memory_mb,
#             stream_to_disk=stream_to_disk
#         )
#         return self._run_async(coro)
    
#     def create_dataset_file(
#         self,
#         aoi: str,
#         expression: str,
#         output: str,
#         in_crs: str = "epsg:4326",
#         res: float = 0.0001,
#         region: str = "aus",
#         to_crs: str = "epsg:4326",
#         overwrite: bool = True,
#         skip_existing: bool = False,
#         non_interactive: bool = True,
#         poll_interval: int = 30,
#         download_path: str = "/home/user/Downloads",
#     ) -> dict:
#         """Create a dataset file using mass stats operations (synchronous version)."""
#         coro = self._async_client.create_dataset_file(
#             aoi=aoi,
#             expression=expression,
#             output=output,
#             in_crs=in_crs,
#             res=res,
#             region=region,
#             to_crs=to_crs,
#             overwrite=overwrite,
#             skip_existing=skip_existing,
#             non_interactive=non_interactive,
#             poll_interval=poll_interval,
#             download_path=download_path,
#         )
#         return self._run_async(coro)
    
#     # Context manager support
#     def __enter__(self):
#         """Context manager entry."""
#         return self
    
#     def __exit__(self, exc_type, exc_val, exc_tb):
#         """Context manager exit."""
#         self.close()
    
#     def __del__(self):
#         """Destructor to ensure session is closed."""
#         if not self._closed:
#             try:
#                 self._cleanup()
#             except Exception:
#                 pass


import asyncio
import concurrent.futures
import threading
import functools
import inspect
from typing import Optional, Dict, Any, Union, TYPE_CHECKING
from geopandas import GeoDataFrame
from shapely.geometry.base import BaseGeometry as ShapelyGeometry
from .async_client import AsyncClient

# Add type checking imports for better IDE support
if TYPE_CHECKING:
    from .endpoints.dataset_management import DatasetManagement
    from .endpoints.user_management import UserManagement
    from .endpoints.mass_stats import MassStats
    from .endpoints.group_management import GroupManagement
    from .endpoints.space_management import SpaceManagement
    from .endpoints.model_management import ModelManagement
    from .endpoints.auth import AuthClient


class SyncWrapper:
    """Generic synchronous wrapper with __dir__ support for runtime autocomplete."""
    
    def __init__(self, async_obj, sync_client):
        self._async_obj = async_obj
        self._sync_client = sync_client
    
    def __dir__(self):
        """Return list of attributes for autocomplete in interactive environments."""
        async_attrs = [attr for attr in dir(self._async_obj) if not attr.startswith('_')]
        wrapper_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
        return list(set(async_attrs + wrapper_attrs))
    
    def __getattr__(self, name):
        """Dynamically wrap any method call to convert async to sync."""
        attr = getattr(self._async_obj, name)
        
        if callable(attr):
            @functools.wraps(attr)
            def sync_wrapper(*args, **kwargs):
                result = attr(*args, **kwargs)
                if hasattr(result, '__await__'):
                    return self._sync_client._run_async(result)
                return result
            return sync_wrapper
        
        return attr


class SyncClient:
    """
    Thread-safe synchronous wrapper for AsyncClient.
    Uses a persistent event loop in a dedicated thread to avoid event loop conflicts.
    """

    # Add explicit type annotations for endpoint managers
    datasets: 'DatasetManagement'
    users: 'UserManagement' 
    mass_stats: 'MassStats'
    groups: 'GroupManagement'
    space: 'SpaceManagement'
    model: 'ModelManagement'
    auth: 'AuthClient'
    
    def __init__(self, url: Optional[str] = None, api_key: Optional[str] = None, verbose: bool = False):
        self._async_client = AsyncClient(url=url, api_key=api_key, verbose=verbose)
        self._context_entered = False
        self._closed = False
        
        # Thread and event loop management
        self._loop = None
        self._thread = None
        self._loop_ready = None
        self._loop_exception = None
        
        # Initialize endpoint managers with proper typing

        self.datasets = SyncWrapper(self._async_client.datasets, self)
        self.users = SyncWrapper(self._async_client.users, self)
        self.mass_stats = SyncWrapper(self._async_client.mass_stats, self)
        self.groups = SyncWrapper(self._async_client.groups, self)
        self.space = SyncWrapper(self._async_client.space, self)
        self.model = SyncWrapper(self._async_client.model, self)
        self.auth = SyncWrapper(self._async_client.auth, self)
        
        import atexit
        atexit.register(self._cleanup)
    
    def _ensure_event_loop(self) -> None:
        """Ensure we have a persistent event loop in a dedicated thread."""
        if self._loop is None or self._loop.is_closed():
            self._loop_ready = threading.Event()
            self._loop_exception = None
            
            def run_loop():
                """Run the event loop in a dedicated thread."""
                try:
                    # Create a new event loop for this thread
                    self._loop = asyncio.new_event_loop()
                    asyncio.set_event_loop(self._loop)
                    
                    # Signal that the loop is ready
                    self._loop_ready.set()
                    
                    # Run the loop forever (until stopped)
                    self._loop.run_forever()
                except Exception as e:
                    self._loop_exception = e
                    self._loop_ready.set()
                finally:
                    # Clean up when the loop stops
                    if self._loop and not self._loop.is_closed():
                        self._loop.close()
            
            # Start the thread
            self._thread = threading.Thread(target=run_loop, daemon=True)
            self._thread.start()
            
            # Wait for the loop to be ready
            self._loop_ready.wait(timeout=10)
            
            if self._loop_exception:
                raise self._loop_exception
            
            if not self._loop_ready.is_set():
                raise RuntimeError("Event loop failed to start within timeout")
    
    def _run_async(self, coro):
        """
        Run async coroutine using persistent event loop.
        This is the core method that makes everything work.
        """
        # Ensure we have an event loop
        self._ensure_event_loop()
        
        if self._loop.is_closed():
            raise RuntimeError("Event loop is closed")
        
        # Create a future to get the result back from the event loop thread
        future = concurrent.futures.Future()
        
        async def run_with_context():
            """Run the coroutine with proper context management."""
            try:
                # Ensure the async client is properly initialized
                await self._ensure_context()
                
                # Run the actual coroutine
                result = await coro
                
                # Set the result on the future
                future.set_result(result)
            except Exception as e:
                # Set the exception on the future
                future.set_exception(e)
        
        # Schedule the coroutine on the persistent event loop
        self._loop.call_soon_threadsafe(
            lambda: asyncio.create_task(run_with_context())
        )

        
        # Wait for the result (with timeout to avoid hanging)
        try:
            return future.result(timeout=300)  # 5 minute timeout
        except concurrent.futures.TimeoutError:
            raise RuntimeError("Async operation timed out after 5 minutes")
    
    async def _ensure_context(self) -> None:
        """Ensure the async client context is entered."""
        if not self._context_entered and not self._closed:
            await self._async_client.__aenter__()
            self._context_entered = True
    
    async def _exit_context(self) -> None:
        """Exit the async client context."""
        if self._context_entered and not self._closed:
            await self._async_client.__aexit__(None, None, None)
            self._context_entered = False
    
    def close(self) -> None:
        """Close the underlying async client session and stop the event loop."""
        if not self._closed:
            if self._loop and not self._loop.is_closed():
                # Schedule cleanup on the event loop
                future = concurrent.futures.Future()
                
                async def cleanup():
                    """Clean up the async client."""
                    try:
                        await self._exit_context()
                        future.set_result(None)
                    except Exception as e:
                        future.set_exception(e)
                
                # Run cleanup
                self._loop.call_soon_threadsafe(
                    lambda: asyncio.create_task(cleanup())
                )
                
                # Wait for cleanup to complete
                try:
                    future.result(timeout=10)
                except:
                    pass  # Ignore cleanup errors
                
                # Stop the event loop
                self._loop.call_soon_threadsafe(self._loop.stop)
                
                # Wait for thread to finish
                if self._thread and self._thread.is_alive():
                    self._thread.join(timeout=5)
            
            self._closed = True
    
    def _cleanup(self) -> None:
        """Internal cleanup method called by atexit."""
        if not self._closed:
            try:
                self.close()
            except Exception:
                pass  # Ignore cleanup errors
    
    def __dir__(self):
        """Return list of attributes for autocomplete in interactive environments."""
        default_attrs = [attr for attr in object.__dir__(self) if not attr.startswith('_')]
        async_client_attrs = [attr for attr in dir(self._async_client) if not attr.startswith('_')]
        endpoint_attrs = ['datasets', 'users', 'mass_stats', 'groups', 'space', 'model', 'auth']
        all_attrs = default_attrs + async_client_attrs + endpoint_attrs
        return list(set(all_attrs))
    
    # Your existing methods with proper type annotations
    def geoquery(
        self,
        expr: str,
        feature: Union[Dict[str, Any], ShapelyGeometry],
        in_crs: str = "epsg:4326",
        out_crs: str = "epsg:4326",
        resolution: int = -1,
        geom_fix: bool = False,
        **kwargs
    ):
        """Compute WCS query for a single geometry (synchronous version)."""
        coro = self._async_client.geoquery(
            expr=expr,
            feature=feature,
            in_crs=in_crs,
            out_crs=out_crs,
            output="netcdf",
            resolution=resolution,
            geom_fix=geom_fix,
            **kwargs
        )
        return self._run_async(coro)

    def zonal_stats(
        self,
        gdf: GeoDataFrame,
        expr: str,
        conc: int = 20,
        in_crs: str = "epsg:4326",
        out_crs: str = "epsg:4326",
        resolution: int = -1,
        geom_fix: bool = False,
        mass_stats: bool = False,
        id_column: Optional[str] = None,
    ) -> GeoDataFrame:

        """
        Compute zonal statistics for all geometries in a GeoDataFrame (synchronous version).
        
        Args:
            gdf (GeoDataFrame): GeoDataFrame containing geometries
            expr (str): Terrakio expression to evaluate, can include spatial aggregations
            conc (int): Number of concurrent requests to make
            in_crs (str): Input coordinate reference system
            out_crs (str): Output coordinate reference system
            resolution (int): Resolution parameter
            geom_fix (bool): Whether to fix the geometry (default False)
            mass_stats (bool): Whether to use mass stats for processing (default False)
            id_column (Optional[str]): Name of the ID column to use (default None)

        Returns:
            geopandas.GeoDataFrame: GeoDataFrame with added columns for results

        Raises:
            ValueError: If concurrency is too high or if data exceeds memory limit without streaming
            APIError: If the API request fails
        """
        coro = self._async_client.zonal_stats(
            gdf=gdf,
            expr=expr,
            conc=conc,
            in_crs=in_crs,
            out_crs=out_crs,
            resolution=resolution,
            geom_fix=geom_fix,
            mass_stats=mass_stats,
            id_column=id_column,
        )
        return self._run_async(coro)
    
    def create_dataset_file(
        self,
        aoi: str,
        expression: str,
        output: str,
        in_crs: str = "epsg:4326",
        res: float = 0.0001,
        region: str = "aus",
        to_crs: str = "epsg:4326",
        overwrite: bool = True,
        skip_existing: bool = False,
        non_interactive: bool = True,
        poll_interval: int = 30,
        download_path: str = "/home/user/Downloads",
    ) -> dict:
        """Create a dataset file using mass stats operations (synchronous version)."""
        coro = self._async_client.create_dataset_file(
            aoi=aoi,
            expression=expression,
            output=output,
            in_crs=in_crs,
            res=res,
            region=region,
            to_crs=to_crs,
            overwrite=overwrite,
            skip_existing=skip_existing,
            non_interactive=non_interactive,
            poll_interval=poll_interval,
            download_path=download_path,
        )
        return self._run_async(coro)

    def geo_queries(
        self,
        queries: list[dict],
        conc: int = 20,
    ) -> Union[float, GeoDataFrame]:
        """
        Execute multiple geo queries concurrently (synchronous version).

        Args:
            queries (list[dict]): List of dictionaries containing query parameters.
                                  Each query must have 'expr', 'feature', and 'in_crs' keys.
            conc (int): Number of concurrent requests to make (default 20, max 100)

        Returns:
            Union[float, geopandas.GeoDataFrame]: 
                - float: Average of all results if results are integers
                - GeoDataFrame: GeoDataFrame with geometry and dataset columns if results are xarray datasets

        Raises:
            ValueError: If queries list is empty, concurrency is too high, or queries are malformed
            APIError: If the API request fails

        Example:
            queries = [
                {
                    'expr': 'WCF.wcf',
                    'feature': {'type': 'Feature', 'geometry': {...}, 'properties': {}},
                    'in_crs': 'epsg:4326'
                },
                {
                    'expr': 'NDVI.ndvi',
                    'feature': {'type': 'Feature', 'geometry': {...}, 'properties': {}},
                    'in_crs': 'epsg:4326'
                }
            ]
            result = client.geo_queries(queries)
        """
        coro = self._async_client.geo_queries(
            queries=queries,
            conc=conc,
        )
        return self._run_async(coro)
    
    # Context manager support
    def __enter__(self) -> 'SyncClient':
        """Context manager entry."""
        return self
    
    def __exit__(self, exc_type, exc_val, exc_tb) -> None:
        """Context manager exit."""
        self.close()
    
    def __del__(self) -> None:
        """Destructor to ensure session is closed."""
        if not self._closed:
            try:
                self._cleanup()
            except Exception:
                pass


import os
import json
from pathlib import Path
from typing import Dict, Any, Optional
import logging
from .exceptions import ConfigurationError

# Default configuration file locations
DEFAULT_CONFIG_FILE = os.path.join(os.environ.get("HOME", ""), ".tkio_config.json")
DEFAULT_API_URL = "https://api.terrak.io"

def read_config_file(config_file: str = DEFAULT_CONFIG_FILE, logger: logging.Logger = None) -> Dict[str, Any]:
    """
    Read and parse the configuration file.
    
    Args:
        config_file: Path to the configuration file
        logger: Logger object to log messages
    Returns:
        Dict[str, Any]: Configuration parameters with additional flags:
                       'is_logged_in': True if user is logged in
                       'user_email': The email of the logged in user
                       'token': Personal token if available
        
    Note:
        This function no longer raises ConfigurationError. Instead, it creates an empty config
        file if one doesn't exist and returns appropriate status flags.
    """
    config_path = Path(os.path.expanduser(config_file))
    # the first circumstance is that the config file does not exist
    # that we need to login before using any of the functions
    # Check if config file exists
    if not config_path.exists():
        # Create an empty config file
        config_path.parent.mkdir(parents=True, exist_ok=True)
        with open(config_path, 'w') as f:
            json.dump({}, f)
        logger.info("No API key found. Please provide an API key to use this client.")
        return {
            'url': DEFAULT_API_URL,
            'key': None,
            'is_logged_in': False,
            'user_email': None,
            'token': None
        }
    
    try:
        # Read the config file
        with open(config_path, 'r') as f:
            config_data = json.load(f)
        
        # Read the config file data
        # Check if config has an API key
        if not config_data or 'TERRAKIO_API_KEY' not in config_data or not config_data.get('TERRAKIO_API_KEY'):
            logger.info("No API key found. Please provide an API key to use this client.")
            return {
                'url': DEFAULT_API_URL,
                'key': None,
                'is_logged_in': False,
                'user_email': None,
                'token': config_data.get('PERSONAL_TOKEN')
            }
        logger.info(f"Currently logged in as: {config_data.get('EMAIL')}")
        # this meanb that we have already logged in to the tkio account
        
        # Convert the JSON config to our expected format
        config = {
            # Always use the default URL, not from config file
            'url': DEFAULT_API_URL,
            'key': config_data.get('TERRAKIO_API_KEY'),
            'is_logged_in': True,
            'user_email': config_data.get('EMAIL'),
            'token': config_data.get('PERSONAL_TOKEN')
        }
        return config
            

    except Exception as e:
        logger.info(f"Error reading config: {e}")
        logger.info("No API key found. Please provide an API key to use this client.")
        return {
            'url': DEFAULT_API_URL,
            'key': None,
            'is_logged_in': False,
            'user_email': None,
            'token': None
        }

def create_default_config(email: str, api_key: str, config_file: str = DEFAULT_CONFIG_FILE) -> None:
    """
    Create a default configuration file in JSON format.
    
    Args:
        email: User email
        api_key: Terrakio API key
        config_file: Path to configuration file
        
    Raises:
        ConfigurationError: If the configuration file can't be created
    """
    config_path = Path(os.path.expanduser(config_file))
    
    # Ensure directory exists
    config_path.parent.mkdir(parents=True, exist_ok=True)
    
    try:
        config_data = {
            "EMAIL": email,
            "TERRAKIO_API_KEY": api_key
        }
        
        with open(config_path, 'w') as f:
            json.dump(config_data, f, indent=2)
            
    except Exception as e:
        raise ConfigurationError(f"Failed to create configuration file: {e}")


# terrakio_core/decorators.py
from functools import wraps
from ..exceptions import ConfigurationError

def require_token(func):
    """Decorator to ensure a token is available before a method can be executed."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check both direct token and client token
        has_token = False
        if hasattr(self, 'token') and self.token:
            has_token = True
        elif hasattr(self, '_client') and hasattr(self._client, 'token') and self._client.token:
            has_token = True
            
        if not has_token:
            raise ConfigurationError("Authentication token required. Please login first.")
        return func(self, *args, **kwargs)
    
    wrapper._is_decorated = True
    return wrapper

def require_api_key(func):
    """Decorator to ensure an API key is available before a method can be executed."""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check both direct key and client key
        has_key = False
        if hasattr(self, 'key') and self.key:
            has_key = True
        elif hasattr(self, '_client') and hasattr(self._client, 'key') and self._client.key:
            has_key = True
            
        if not has_key:
            raise ConfigurationError("API key required. Please provide an API key or login first.")
        return func(self, *args, **kwargs)
    
    wrapper._is_decorated = True
    return wrapper

def require_auth(func):
    """Decorator that requires either a token OR an API key"""
    @wraps(func)
    def wrapper(self, *args, **kwargs):
        # Check both direct auth and client auth
        has_token = (hasattr(self, 'token') and self.token) or \
                   (hasattr(self, '_client') and hasattr(self._client, 'token') and self._client.token)
        has_api_key = (hasattr(self, 'key') and self.key) or \
                     (hasattr(self, '_client') and hasattr(self._client, 'key') and self._client.key)
        
        if not has_token and not has_api_key:
            raise ConfigurationError(
                "Authentication required. Please provide either an API key or login to get a token."
            )
        return func(self, *args, **kwargs)
    
    wrapper._is_decorated = True
    return wrapper


import aiohttp
import asyncio
import json
import pandas as pd
import xarray as xr
from io import BytesIO
from typing import Optional, Dict, Any, Union
from geopandas import GeoDataFrame
from shapely.geometry.base import BaseGeometry as ShapelyGeometry
from shapely.geometry import mapping
from .client import BaseClient
from .exceptions import APIError
from .endpoints.dataset_management import DatasetManagement
from .endpoints.user_management import UserManagement
from .endpoints.mass_stats import MassStats
from .endpoints.group_management import GroupManagement
from .endpoints.space_management import SpaceManagement
from .endpoints.model_management import ModelManagement
from .endpoints.auth import AuthClient
from .convenience_functions.convenience_functions import zonal_stats as _zonal_stats, create_dataset_file as _create_dataset_file, request_geoquery_list as _request_geoquery_list

class AsyncClient(BaseClient):
    def __init__(self, url: Optional[str] = None, api_key: Optional[str] = None, verbose: bool = False, session: Optional[aiohttp.ClientSession] = None):
        super().__init__(url, api_key, verbose)
        self.datasets = DatasetManagement(self)
        self.users = UserManagement(self)
        self.mass_stats = MassStats(self)
        self.groups = GroupManagement(self)
        self.space = SpaceManagement(self)
        self.model = ModelManagement(self)
        self.auth = AuthClient(self)
        self._session = session
        self._owns_session = session is None

    async def _terrakio_request(self, method: str, endpoint: str, **kwargs):
        if self.session is None:
            headers = {
                'Content-Type': 'application/json',
                'x-api-key': self.key,
                'Authorization': self.token
            }
            clean_headers = {k: v for k, v in headers.items() if v is not None}
            async with aiohttp.ClientSession(headers=clean_headers, timeout=aiohttp.ClientTimeout(total=self.timeout)) as session:
                return await self._make_request_with_retry(session, method, endpoint, **kwargs)
        else:
            return await self._make_request_with_retry(self._session, method, endpoint, **kwargs)

    async def _make_request_with_retry(self, session: aiohttp.ClientSession, method: str, endpoint: str, **kwargs) -> Dict[Any, Any]:
        url = f"{self.url}/{endpoint.lstrip('/')}"    
        last_exception = None
        for attempt in range(self.retry + 1):
            try:
                async with session.request(method, url, **kwargs) as response:
                    if not response.ok and self._should_retry(response.status, attempt):
                        self.logger.info(f"Request failed (attempt {attempt+1}/{self.retry+1}): {response.status}. Retrying...")
                        continue
                    if not response.ok:
                        error_msg = f"API request failed: {response.status} {response.reason}"
                        try:
                            error_data = await response.json()
                            if "detail" in error_data:
                                error_msg += f" - {error_data['detail']}"
                        except:
                            pass
                        raise APIError(error_msg, status_code=response.status)
                    return await self._parse_response(response)
                    
            except aiohttp.ClientError as e:
                last_exception = e
                if attempt < self.retry:
                    self.logger.info(f"Networking error (attempt {attempt+1}/{self.retry+1}): {e}. Retrying...")
                    continue
                else:
                    break
        
        raise APIError(f"Networking error, request failed after {self.retry+1} attempts: {last_exception}", status_code=None)
    
    def _should_retry(self, status_code: int, attempt: int) -> bool:
        """Determine if the request should be retried based on status code."""
        if attempt >= self.retry:
            return False
        elif status_code in [408, 502, 503, 504]:
            return True
        else:
            return False

    async def _parse_response(self, response) -> Any:
        """Parse response based on content type."""
        content_type = response.headers.get('content-type', '').lower()
        content = await response.read()
        if 'json' in content_type:
            return json.loads(content.decode('utf-8'))
        elif 'csv' in content_type:
            return pd.read_csv(BytesIO(content))
        elif 'image/' in content_type:
            return content
        elif 'text' in content_type:
            return content.decode('utf-8')
        else:
            try:
                return xr.open_dataset(BytesIO(content))
            except:
                raise APIError(f"Unknown response format: {content_type}", status_code=response.status)

    async def _regular_request(self, method: str, endpoint: str, **kwargs):
        url = endpoint.lstrip('/')
        
        if self._session is None:
            async with aiohttp.ClientSession() as session:
                try:
                    async with session.request(method, url, **kwargs) as response:
                        response.raise_for_status()
                        
                        content = await response.read()
                        
                        return type('Response', (), {
                            'status': response.status,
                            'content': content,
                            'text': lambda: content.decode('utf-8'),
                            'json': lambda: json.loads(content.decode('utf-8'))
                        })()
                except aiohttp.ClientError as e:
                    raise APIError(f"Request failed: {e}")
        else:
            try:
                async with self._session.request(method, url, **kwargs) as response:
                    response.raise_for_status()
                    content = await response.read()
                    
                    return type('Response', (), {
                        'status': response.status,
                        'content': content,
                        'text': lambda: content.decode('utf-8'),
                        'json': lambda: json.loads(content.decode('utf-8'))
                    })()
            except aiohttp.ClientError as e:
                raise APIError(f"Request failed: {e}")
        
    async def geoquery(
        self,
        expr: str,
        feature: Union[Dict[str, Any], ShapelyGeometry],
        in_crs: str = "epsg:4326",
        out_crs: str = "epsg:4326",
        resolution: int = -1,
        geom_fix: bool = False,
        validated: bool = True,
        **kwargs
    ):
        """
        Compute WCS query for a single geometry.

        Args:
            expr (str): The WCS expression to evaluate
            feature (Union[Dict[str, Any], ShapelyGeometry]): The geographic feature
            in_crs (str): Input coordinate reference system
            out_crs (str): Output coordinate reference system
            resolution (int): Resolution parameter
            geom_fix (bool): Whether to fix the geometry (default False)
            validated (bool): Whether to use validated data (default True)
            **kwargs: Additional parameters to pass to the WCS request
            
        Returns:
            Union[pd.DataFrame, xr.Dataset, bytes]: The response data in the requested format

        Raises:
            APIError: If the API request fails
        """
        if hasattr(feature, 'is_valid'):
            feature = {
                "type": "Feature",
                "geometry": mapping(feature),
                "properties": {}
            }
        payload = {
            "feature": feature,
            "in_crs": in_crs,
            "out_crs": out_crs,
            "output": "netcdf",
            "resolution": resolution,
            "expr": expr,
            "buffer": geom_fix,
            "validated": validated,
            **kwargs
        }
        result = await self._terrakio_request("POST", "geoquery", json=payload)

        return result

    async def zonal_stats(
            self,
            gdf: GeoDataFrame,
            expr: str,
            conc: int = 20,
            in_crs: str = "epsg:4326",
            out_crs: str = "epsg:4326",
            resolution: int = -1,
            geom_fix: bool = False,
            mass_stats: bool = False,
            id_column: Optional[str] = None,
    ):
        """
        Compute zonal statistics for all geometries in a GeoDataFrame.

        Args:
            gdf (GeoDataFrame): GeoDataFrame containing geometries
            expr (str): Terrakio expression to evaluate, can include spatial aggregations
            conc (int): Number of concurrent requests to make
            in_crs (str): Input coordinate reference system
            out_crs (str): Output coordinate reference system
            resolution (int): Resolution parameter
            geom_fix (bool): Whether to fix the geometry (default False)
            mass_stats (bool): Whether to use mass stats for processing (default False)
            id_column (Optional[str]): Name of the ID column to use (default None)

        Returns:
            geopandas.GeoDataFrame: GeoDataFrame with added columns for results

        Raises:
            ValueError: If concurrency is too high or if data exceeds memory limit without streaming
            APIError: If the API request fails
        """
        return await _zonal_stats(
            client=self,
            gdf=gdf,
            expr=expr,
            conc=conc,
            in_crs=in_crs,
            out_crs=out_crs,
            resolution=resolution,
            geom_fix=geom_fix,
            mass_stats=mass_stats,
            id_column=id_column,
        )

    async def create_dataset_file(
        self,
        aoi: str,
        expression: str,
        output: str,
        in_crs: str = "epsg:4326",
        res: float = 0.0001,
        region: str = "aus",
        to_crs: str = "epsg:4326",
        overwrite: bool = True,
        skip_existing: bool = False,
        non_interactive: bool = True,
        poll_interval: int = 30,
        download_path: str = "/home/user/Downloads",
    ) -> dict:
        """
        Create a dataset file using mass stats operations.

        Args:
            aoi (str): Area of interest
            expression (str): Terrakio expression to evaluate
            output (str): Output format
            in_crs (str): Input coordinate reference system (default "epsg:4326")
            res (float): Resolution (default 0.0001)
            region (str): Region (default "aus")
            to_crs (str): Target coordinate reference system (default "epsg:4326")
            overwrite (bool): Whether to overwrite existing files (default True)
            skip_existing (bool): Whether to skip existing files (default False)
            non_interactive (bool): Whether to run non-interactively (default True)
            poll_interval (int): Polling interval in seconds (default 30)
            download_path (str): Download path (default "/home/user/Downloads")

        Returns:
            dict: Dictionary containing generation_task_id and combine_task_id

        Raises:
            ConfigurationError: If mass stats client is not properly configured
            RuntimeError: If job fails
        """
        return await _create_dataset_file(
            client=self,
            aoi=aoi,
            expression=expression,
            output=output,
            in_crs=in_crs,
            res=res,
            region=region,
            to_crs=to_crs,
            overwrite=overwrite,
            skip_existing=skip_existing,
            non_interactive=non_interactive,
            poll_interval=poll_interval,
            download_path=download_path,
        )

    async def geo_queries(
        self,
        queries: list[dict],
        conc: int = 20,
    ):
        """
        Execute multiple geo queries concurrently.

        Args:
            queries (list[dict]): List of dictionaries containing query parameters.
                                  Each query must have 'expr', 'feature', and 'in_crs' keys.
            conc (int): Number of concurrent requests to make (default 20, max 100)

        Returns:
            Union[float, geopandas.GeoDataFrame]: 
                - float: Average of all results if results are integers
                - GeoDataFrame: GeoDataFrame with geometry and dataset columns if results are xarray datasets

        Raises:
            ValueError: If queries list is empty, concurrency is too high, or queries are malformed
            APIError: If the API request fails

        Example:
            queries = [
                {
                    'expr': 'WCF.wcf',
                    'feature': {'type': 'Feature', 'geometry': {...}, 'properties': {}},
                    'in_crs': 'epsg:4326'
                },
                {
                    'expr': 'NDVI.ndvi',
                    'feature': {'type': 'Feature', 'geometry': {...}, 'properties': {}},
                    'in_crs': 'epsg:4326'
                }
            ]
            result = await client.geo_queries(queries)
        """
        return await _request_geoquery_list(
            client=self,
            quries=queries,  # Note: keeping original parameter name for compatibility
            conc=conc,
        )

    async def __aenter__(self):
        if self._session is None:
            headers = {
                'Content-Type': 'application/json',
                'x-api-key': self.key,
                'Authorization': self.token
            }
            clean_headers = {k: v for k, v in headers.items() if v is not None}
            self._session = aiohttp.ClientSession(
                headers=clean_headers,
                timeout=aiohttp.ClientTimeout(total=self.timeout)
            )
        return self

    async def __aexit__(self, exc_type, exc_val, exc_tb):
        if self._owns_session and self._session:
            await self._session.close()
            self._session = None

    async def close(self):
        if self._owns_session and self._session:
            await self._session.close()
            self._session = None


from typing import Optional
import logging
import warnings
from terrakio_core.config import read_config_file, DEFAULT_CONFIG_FILE
from abc import abstractmethod
import xarray as xr


class BaseClient():
    def __init__(self, url: Optional[str] = None, api_key: Optional[str] = None, verbose: bool = False):
        self.verbose = verbose
        self.logger = logging.getLogger("terrakio")
        if verbose:
            self.logger.setLevel(logging.INFO)
        else:
            self.logger.setLevel(logging.WARNING)

        self.timeout = 300
        self.retry = 3

        self.session = None

        self.url = url
        self.key = api_key

        config = read_config_file(DEFAULT_CONFIG_FILE, logger=self.logger)
                
        if self.url is None:
            self.url = config.get('url')
                
        if self.key is None:
            self.key = config.get('key')

        self.token = config.get('token')
        
    #     # Apply xarray printing fix to prevent crashes with GeoDataFrames
    #     self._apply_xarray_fix()
    
    # def _apply_xarray_fix(self):
    #     """
    #     Apply xarray printing fix to prevent crashes when GeoDataFrames contain xarray objects.
    #     This fix is applied automatically when the client is initialized.
    #     """
    #     try:
            
    #         # Check if fix is already applied globally
    #         if hasattr(xr.DataArray, '_terrakio_fix_applied'):
    #             if self.verbose:
    #                 self.logger.info("xarray printing fix already applied")
    #             return
            
    #         # Store original methods for potential restoration
    #         if not hasattr(xr.DataArray, '_original_iter'):
    #             xr.DataArray._original_iter = xr.DataArray.__iter__
    #             xr.Dataset._original_iter = xr.Dataset.__iter__
            
    #         # Define safe iteration methods that prevent pandas from iterating
    #         # but leave __repr__ and __str__ untouched for normal xarray printing
    #         def safe_dataarray_iter(self):
    #             # Return infinite iterator that always yields the same safe value
    #             name = getattr(self, 'name', None) or 'unnamed'
    #             shape_str = 'x'.join(map(str, self.shape)) if hasattr(self, 'shape') else 'unknown'
    #             placeholder = f"<DataArray '{name}' {shape_str}>"
    #             while True:
    #                 yield placeholder
            
    #         def safe_dataset_iter(self):
    #             # Return infinite iterator that always yields the same safe value
    #             num_vars = len(self.data_vars) if hasattr(self, 'data_vars') else 0
    #             num_dims = len(self.dims) if hasattr(self, 'dims') else 0
    #             placeholder = f"<Dataset: {num_vars} vars, {num_dims} dims>"
    #             while True:
    #                 yield placeholder
            
    #         # Apply only the iteration fix - leave __repr__ and __str__ untouched
    #         xr.DataArray.__iter__ = safe_dataarray_iter
    #         xr.Dataset.__iter__ = safe_dataset_iter
            
    #         # Mark as applied to avoid duplicate applications
    #         xr.DataArray._terrakio_fix_applied = True
    #         xr.Dataset._terrakio_fix_applied = True
            
    #         if self.verbose:
    #             self.logger.info("xarray iteration fix applied - GeoDataFrames with xarray objects will print safely, direct xarray printing unchanged")
                
    #     except ImportError:
    #         # xarray not installed, skip the fix
    #         if self.verbose:
    #             self.logger.info("xarray not installed, skipping printing fix")
    #     except Exception as e:
    #         # Log warning but don't fail initialization
    #         warning_msg = f"Failed to apply xarray printing fix: {e}"
    #         warnings.warn(warning_msg)
    #         if self.verbose:
    #             self.logger.warning(warning_msg)
    
    # def restore_xarray_printing(self):
    #     """
    #     Restore original xarray printing behavior.
    #     Call this method if you want to see full xarray representations again.
    #     """
    #     try:
    #         import xarray as xr
            
    #         if hasattr(xr.DataArray, '_original_iter'):
    #             xr.DataArray.__iter__ = xr.DataArray._original_iter
    #             xr.Dataset.__iter__ = xr.Dataset._original_iter
                
    #             # Remove the fix markers
    #             if hasattr(xr.DataArray, '_terrakio_fix_applied'):
    #                 delattr(xr.DataArray, '_terrakio_fix_applied')
    #             if hasattr(xr.Dataset, '_terrakio_fix_applied'):
    #                 delattr(xr.Dataset, '_terrakio_fix_applied')
                
    #             if self.verbose:
    #                 self.logger.info("Original xarray iteration behavior restored")
    #         else:
    #             if self.verbose:
    #                 self.logger.info("No xarray fix to restore")
                    
    #     except ImportError:
    #         if self.verbose:
    #             self.logger.info("xarray not available")
    #     except Exception as e:
    #         warning_msg = f"Failed to restore xarray printing: {e}"
    #         warnings.warn(warning_msg)
    #         if self.verbose:
    #             self.logger.warning(warning_msg)
    
    # @abstractmethod
    # def _setup_session(self):
    #     """Initialize the HTTP session - implemented by sync/async clients"""
    #     pass


from typing import Dict, Any, List, Optional
from ..helper.decorators import require_token, require_api_key, require_auth

class UserManagement:
    def __init__(self, client):
        self._client = client

    @require_api_key
    def get_user_by_id(self, id: str) -> Dict[str, Any]:
        """
        Get user by ID.

        Args:
            user_id: User ID
        
        Returns:
            User information
            
        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", f"admin/users/{id}")
    
    @require_api_key
    def get_user_by_email(self, email: str) -> Dict[str, Any]:
        """
        Get user by email.

        Args:
            email: User email
        
        Returns:
            User information
            
        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", f"admin/users/email/{email}")
    
    @require_api_key
    def list_users(self, substring: Optional[str] = None, uid: bool = False) -> List[Dict[str, Any]]:
        """
        List users, optionally filtering by a substring.
        
        Args:
            substring: Optional substring to filter users
            uid: If True, includes the user ID in the response (default: False)
        
        Returns:
            List of users
            
        Raises:
            APIError: If the API request fails
        """
        params = {"uid": str(uid).lower()}
        if substring:
            params['substring'] = substring
        return self._client._terrakio_request("GET", "admin/users", params=params)
    
    @require_api_key
    def edit_user(
        self,
        uid: str,
        email: Optional[str] = None,
        role: Optional[str] = None,
        apiKey: Optional[str] = None,
        groups: Optional[List[str]] = None,
        quota: Optional[int] = None
    ) -> Dict[str, Any]:
        """
        Edit user info. Only provided fields will be updated.
        
        Args:
            uid: User ID
            email: New user email
            role: New user role
            apiKey: New API key
            groups: New list of groups
            quota: New quota
        
        Returns:
            Updated user information
            
        Raises:
            APIError: If the API request fails
        """
        payload = {"uid": uid}
        payload_mapping = {
            "email": email,
            "role": role,
            "apiKey": apiKey,
            "groups": groups,
            "quota": quota
        }
        for key, value in payload_mapping.items():
            if value is not None:
                payload[key] = value
        return self._client._terrakio_request("PATCH", "admin/users", json=payload)
    
    @require_api_key
    def reset_quota(self, email: str, quota: Optional[int] = None) -> Dict[str, Any]:
        """
        Reset the quota for a user by email.
        
        Args:
            email: The user's email (required)
            quota: The new quota value (optional)
            
        Returns:
            API response as a dictionary
        """
        payload = {"email": email}
        if quota is not None:
            payload["quota"] = quota
        return self._client._terrakio_request("PATCH", f"admin/users/reset_quota/{email}", json=payload)
    
    @require_api_key
    def delete_user(self, uid: str) -> Dict[str, Any]:
        """
        Delete a user by UID.

        Args:
            uid: The user's UID (required)
            
        Returns:
            API response as a dictionary
            
        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("DELETE", f"admin/users/{uid}")


from typing import Dict, Any
from ..helper.decorators import require_token, require_api_key, require_auth
class BaseGroupManagement:
    """Base class with common group management methods."""
    def __init__(self, client):
        self._client = client

    @require_api_key
    def get_group_datasets(self, group: str, collection: str = "terrakio-datasets") -> Dict[str, Any]:
        """
        Get datasets of a group.

        Args:
            group: Name of the group
            collection: Name of the collection (default is "terrakio-datasets")

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        params = {"collection": collection}
        return self._client._terrakio_request("GET", f"/groups/{group}/datasets", params=params)

    @require_api_key
    def add_user_to_group(self, group: str, emails: list[str]) -> Dict[str, Any]:
        """
        Add a user to a group.

        Args:
            group: Name of the group
            email: List of user emails to add to the group

        Returns:
            API response data
        """
        payload = {"emails": emails}
        return self._client._terrakio_request("POST", f"/groups/{group}/users", json = payload)

    @require_api_key
    def add_group_to_dataset(self, dataset: str, id: str) -> Dict[str, Any]:
        """
        Add a group to a dataset.

        Args:
            dataset: Name of the dataset
            id: Group ID

        Returns:
            API response data
        """
        payload = {"id": id}
        return self._client._terrakio_request("POST", f"/datasets/{dataset}/groups", json = payload)

    @require_api_key
    def add_user_to_dataset(self, dataset: str, emails: list[str]) -> Dict[str, Any]:
        """
        Add a user to a dataset.

        Args:
            dataset: Name of the dataset
            email: List of user emails to add to the dataset

        Returns:
            API response data
        """
        payload = {"emails": emails}
        return self._client._terrakio_request("POST", f"/datasets/{dataset}/share", json = payload)

    @require_api_key
    def remove_user_from_group(self, group: str, emails: list[str]) -> Dict[str, Any]:
        """
        Remove a user from a group.

        Args:
            group: Name of the group
            email: List of user emails to remove from the group

        Returns:
            API response data
        """
        payload = {"emails": emails}
        return self._client._terrakio_request("DELETE", f"/groups/{group}/users", json = payload)
    
    @require_api_key
    def remove_user_from_dataset(self, dataset: str, emails: list[str]) -> Dict[str, Any]:
        """
        Remove a user from a dataset.

        Args:
            dataset: Name of the dataset
            email: List of user emails to remove from the dataset

        Returns:
            API response data
        """
        payload = {"emails": emails}
        return self._client._terrakio_request("PATCH", f"/datasets/{dataset}/share", json = payload)

class GroupManagement(BaseGroupManagement):
    """Normal user group management with regular user permissions."""
    @require_api_key
    def list_groups(self) -> Dict[str, Any]:
        """
        List all groups.

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", "/groups")
    
    @require_api_key
    def get_group(self, group: str) -> Dict[str, Any]:
        """
        Get a group.

        Args:
            group: Name of the group

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", f"/groups/{group}")

    @require_api_key
    def create_group(self, name: str) -> Dict[str, Any]:
        """
        Create a group

        Args:
            name: Name of the group

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        payload = {"name": name}
        return self._client._terrakio_request("POST", "/groups", json = payload)

    @require_api_key
    def delete_group(self, group: str) -> Dict[str, Any]:
        """
        Delete a group.

        Args:
            group: Name of the group to delete

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("DELETE", f"/groups/{group}")

class AdminGroupManagement(BaseGroupManagement):
    """Admin user group management with admin permissions."""
    @require_api_key
    def list_groups(self) -> Dict[str, Any]:
        """
        List all groups.

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", "/admin/groups")
    
    @require_api_key
    def get_group(self, group: str) -> Dict[str, Any]:
        """
        Get a group.

        Args:
            group: Name of the group

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", f"/admin/groups/{group}")
    
    @require_api_key
    def create_group(self, name: str, owner: str) -> Dict[str, Any]:
        """
        Create a group.

        Args:
            name: Name of the group
            owner: User email of the owner

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        payload = {"name": name, "owner": owner}
        return self._client._terrakio_request("POST", f"/admin/groups", json=payload)
    
    @require_api_key
    def delete_group(self, group: str) -> Dict[str, Any]:
        """
        Delete a group.

        Args:
            group: Name of the group to delete

        Returns:
            API response data

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("DELETE", f"/admin/groups/{group}")


# Standard library imports
import ast
import json
import textwrap
import time
from io import BytesIO
from typing import Optional, Tuple
import onnxruntime as ort

# Internal imports
from ..helper.decorators import require_api_key

# Optional dependency flags
TORCH_AVAILABLE = False
SKL2ONNX_AVAILABLE = False

# PyTorch imports
try:
    import torch
    TORCH_AVAILABLE = True
except ImportError:
    torch = None

# Scikit-learn and ONNX conversion imports
try:
    from sklearn.base import BaseEstimator
    from skl2onnx import convert_sklearn
    from skl2onnx.common.data_types import FloatTensorType
    SKL2ONNX_AVAILABLE = True
except ImportError:
    BaseEstimator = None
    convert_sklearn = None
    FloatTensorType = None

class ModelManagement:
    def __init__(self, client):
        self._client = client

    @require_api_key
    async def generate_ai_dataset(
        self,
        name: str,
        aoi_geojson: str,
        expression_x: str,
        filter_x_rate: float,
        filter_y_rate: float,
        samples: int,
        tile_size: int,
        expression_y: str = "skip",
        filter_x: str = "skip",
        filter_y: str = "skip",
        crs: str = "epsg:4326",
        res: float = 0.001,
        region: str = None,
        bucket: str = None,
        start_year: int = None,
        end_year: int = None,
    ) -> dict:
        """
        Generate an AI dataset using specified parameters.

        Args:
            name (str): Name of the dataset to generate
            aoi_geojson (str): Path to GeoJSON file containing area of interest
            expression_x (str): Expression for X variable (e.g. "MSWX.air_temperature@(year=2021, month=1)")
            filter_x (str): Filter for X variable (e.g. "MSWX.air_temperature@(year=2021, month=1)")
            filter_x_rate (float): Filter rate for X variable (e.g. 0.5)
            expression_y (str): Expression for Y variable with {year} placeholder
            filter_y (str): Filter for Y variable (e.g. "MSWX.air_temperature@(year=2021, month=1)")
            filter_y_rate (float): Filter rate for Y variable (e.g. 0.5)
            samples (int): Number of samples to generate
            tile_size (int): Size of tiles in degrees
            crs (str, optional): Coordinate reference system. Defaults to "epsg:4326"
            res (float, optional): Resolution in degrees. Defaults to 0.001
            region (str, optional): Region code. Defaults to None
            bucket (str, optional): Bucket name. Defaults to None
            start_year (int, optional): Start year for data generation. Required if end_year provided
            end_year (int, optional): End year for data generation. Required if start_year provided

        Returns:
            dict: Response from the AI dataset generation API

        Raises:
            APIError: If the API request fails
        """
        # Build config for expressions and filters
        config = {
            "expressions": [{"expr": expression_x, "res": res, "prefix": "x"}],
            "filters": []
        }

        if expression_y != "skip":
            config["expressions"].append({"expr": expression_y, "res": res, "prefix": "y"})

        if filter_x != "skip":
            config["filters"].append({"expr": filter_x, "res": res, "rate": filter_x_rate})
        if filter_y != "skip":
            config["filters"].append({"expr": filter_y, "res": res, "rate": filter_y_rate})

        # Replace year placeholders if start_year is provided
        if start_year is not None:
            expression_x = expression_x.replace("{year}", str(start_year))
            if expression_y != "skip":
                expression_y = expression_y.replace("{year}", str(start_year))
            if filter_x != "skip":
                filter_x = filter_x.replace("{year}", str(start_year))
            if filter_y != "skip":
                filter_y = filter_y.replace("{year}", str(start_year))

        # Load AOI GeoJSON
        with open(aoi_geojson, 'r') as f:
            aoi_data = json.load(f)

        task_response = await self._client.mass_stats.random_sample(
            name=name,
            config=config,
            aoi=aoi_data,
            samples=samples,
            year_range=[start_year, end_year],
            crs=crs,
            tile_size=tile_size,
            res=res,
            region=region,
            output="netcdf",
            server=self._client.url,
            bucket=bucket,
            overwrite=True
        )
        task_id = task_response["task_id"]

        while True:
            result = await self._client.mass_stats.track_job(ids=[task_id])
            status = result[task_id]['status']
            completed = result[task_id].get('completed', 0)
            total = result[task_id].get('total', 1)
            
            progress = completed / total if total > 0 else 0
            bar_length = 50
            filled_length = int(bar_length * progress)
            bar = '█' * filled_length + '░' * (bar_length - filled_length)
            percentage = progress * 100
            
            self._client.logger.info(f"Job status: {status} [{bar}] {percentage:.1f}% ({completed}/{total})")   

            if status == "Completed":
                self._client.logger.info("Job completed successfully!")
                break
            elif status == "Error":
                self._client.logger.info("Job encountered an error")
                raise Exception(f"Job {task_id} encountered an error")
            
            time.sleep(5)

        task_id = await self._client.mass_stats.start_job(task_id)
        return task_id
    
    @require_api_key
    async def _get_url_for_upload_model_and_script(self, expression: str, model_name: str, script_name: str) -> str:
        """
        Get the url for the upload of the model
        Args:
            expression: The expression to use for the upload(for deciding which bucket to upload to)
            model_name: The name of the model to upload
            script_name: The name of the script to upload
        Returns:
            The url for the upload of the model
        """
        payload = {
            "model_name": model_name,
            "expression": expression,
            "script_name": script_name
        }
        return await self._client._terrakio_request("POST", "models/upload", json=payload)

    async def _upload_model_to_url(self, upload_model_url: str, model: bytes):
        """
        Upload a model to a given URL.
        Args:
            model_url: The url to upload the model to
            model: The model to upload

        Returns:
            The response from the server
        """
        headers = {
            "Content-Type": "application/octet-stream",
            "Content-Length": str(len(model))
        }
        response = await self._client._regular_request("PUT", endpoint = upload_model_url, data=model, headers=headers)
        return response
    
    @require_api_key
    async def _upload_script_to_url(self, upload_script_url: str, script_content: str):
        """
            Upload the generated script to the url
            Args:
                url: Url for the upload of the script
                script_content: Content of the script
            returns:
                None
        """
        script_bytes = script_content.encode('utf-8')
        headers = {
            "Content-Type": "text/x-python",
            "Content-Length": str(len(script_bytes))
        }
        response = await self._client._regular_request("PUT", endpoint=upload_script_url, data=script_bytes, headers=headers)
        return response

    @require_api_key
    async def _upload_model_and_script(self, model, model_name: str, script_name: str, input_expression: str, input_shape: Tuple[int, ...] = None, processing_script_path: Optional[str] = None, model_type: Optional[str] = None):
        """
        Upload a model and script to the bucket
        Args:
            model: The model object (PyTorch model or scikit-learn model)
            model_name: Name for the model (without extension)
            script_name: Name for the script (without extension)
            input_expression: Input expression for the dataset
            input_shape: Shape of input data for ONNX conversion (required for PyTorch models)
            processing_script_path: Path to the processing script, if not provided, no processing will be done
            model_type: The type of the model we want to upload
        Raises:
            APIError: If the API request fails
            ValueError: If model type is not supported or input_shape is missing for PyTorch models

        Returns:
            bucket_name: Name of the bucket where the model is stored
        """
        response = await self._get_url_for_upload_model_and_script(expression = input_expression, model_name = model_name, script_name = script_name)
        model_url, script_url, bucket_name = response.get("model_upload_url"), response.get("script_upload_url"), response.get("bucket_name")
        if not model_url or not script_url:
            raise ValueError("No url returned from the server for the upload process")
        try:
            model_in_onnx_bytes, model_type = self._convert_model_to_onnx(model = model, input_shape = input_shape, model_type = model_type)
            if model_type == "neural_network":
                script_content = await self._generate_cnn_script(bucket_name = bucket_name, virtual_dataset_name = model_name, virtual_product_name = script_name, processing_script_path = processing_script_path)
            elif model_type == "random_forest":
                script_content = await self._generate_random_forest_script(bucket_name = bucket_name, virtual_dataset_name = model_name, virtual_product_name = script_name, processing_script_path = processing_script_path)
            else:
                raise ValueError(f"Unsupported model type: {model_type}. Supported types: neural_network, random_forest")
            script_upload_response = await self._upload_script_to_url( upload_script_url = script_url, script_content = script_content)
            if script_upload_response.status not in [200, 201, 204]:
                self._client.logger.error(f"Script upload error: {script_upload_response.text()}")
                raise Exception(f"Failed to upload script: {script_upload_response.text()}")
            model_upload_response = await self._upload_model_to_url(upload_model_url = model_url, model = model_in_onnx_bytes)
            if model_upload_response.status not in [200, 201, 204]:
                self._client.logger.error(f"Model upload error: {model_upload_response.text()}")
                raise Exception(f"Failed to upload model: {model_upload_response.text()}")
        except Exception as e:
            raise Exception(f"Error uploading model: {e}")
        self._client.logger.info(f"Model and Script uploaded successfully to {model_url}")
        return bucket_name

    @require_api_key
    async def upload_and_deploy_model(self, model, virtual_dataset_name: str, virtual_product_name: str, input_expression: str, dates_iso8601: list, input_shape: Tuple[int, ...] = None, processing_script_path: Optional[str] = None, model_type: Optional[str] = None):
        """
        Upload a model to the bucket and deploy it.
        Args:
            model: The model object (PyTorch model or scikit-learn model)
            virtual_dataset_name: Name for the virtual dataset (without extension)
            virtual_product_name: Product name for the inference
            input_expression: Input expression for the dataset
            dates_iso8601: List of dates in ISO8601 format
            input_shape: Shape of input data for ONNX conversion (required for PyTorch models)
            processing_script_path: Path to the processing script, if not provided, no processing will be done
            model_type: The type of the model we want to upload

        Raises:
            APIError: If the API request fails
            ValueError: If model type is not supported or input_shape is missing for PyTorch models
            ImportError: If required libraries (torch or skl2onnx) are not installed

        Returns:
            None
        """
        bucket_name = await self._upload_model_and_script(model=model, model_name=virtual_dataset_name, script_name= virtual_product_name, input_shape=input_shape, input_expression=input_expression, processing_script_path=processing_script_path, model_type= model_type)
        user_info = await self._client.auth.get_user_info()
        uid = user_info["uid"]
        await self._client.datasets.create_dataset(
            name=virtual_dataset_name,
            collection="terrakio-datasets",
            products=[virtual_product_name],
            path=f"gs://{bucket_name}/{uid}/virtual_datasets/{virtual_dataset_name}/inference_scripts",
            input=input_expression,
            dates_iso8601=dates_iso8601,
            padding=0
        )
    
    @require_api_key
    async def _generate_random_forest_script(self, bucket_name: str, virtual_dataset_name: str, virtual_product_name: str, processing_script_path: Optional[str] = None) -> str:
        """
        Generate Python inference script for the Random Forest model.
        
        Args:
            bucket_name: Name of the bucket where the model is stored
            virtual_dataset_name: Name of the virtual dataset and the model
            virtual_product_name: Name of the virtual product
            processing_script_path: Path to the processing script, if not provided, no processing will be done
            
        Returns:
            str: Generated Python script content
        """
        user_info = await self._client.auth.get_user_info()
        uid = user_info["uid"]
        preprocessing_code, postprocessing_code = None, None

        if processing_script_path:
            try:
                preprocessing_code, postprocessing_code = self._parse_processing_script(processing_script_path)
                if preprocessing_code:
                    self._client.logger.info(f"Using custom preprocessing from: {processing_script_path}")
                if postprocessing_code:
                    self._client.logger.info(f"Using custom postprocessing from: {processing_script_path}")
                if not preprocessing_code and not postprocessing_code:
                    self._client.logger.warning(f"No preprocessing or postprocessing functions found in {processing_script_path}")
                    self._client.logger.info("Deployment will continue without custom processing")
            except Exception as e:
                raise ValueError(f"Failed to load processing script: {str(e)}")     
   
        preprocessing_section = ""
        if preprocessing_code and preprocessing_code.strip():
            clean_preprocessing = textwrap.dedent(preprocessing_code)
            preprocessing_section = textwrap.indent(clean_preprocessing, '    ')
        
        postprocessing_section = ""
        if postprocessing_code and postprocessing_code.strip():
            clean_postprocessing = textwrap.dedent(postprocessing_code)
            postprocessing_section = textwrap.indent(clean_postprocessing, '    ')

        script_lines = [
            "import logging",
            "from io import BytesIO",
            "import numpy as np",
            "import pandas as pd",
            "import xarray as xr",
            "from google.cloud import storage",
            "from onnxruntime import InferenceSession",
            "from typing import Tuple",
            "",
            "logging.basicConfig(",
            "    level=logging.INFO",
            ")",
            "",
        ]
        
        if preprocessing_section:
            script_lines.extend([
                "def validate_preprocessing_output(data_arrays):",
                "    \"\"\"",
                "    Validate preprocessing output coordinates and data type.",
                "    ",
                "    Args:",
                "        data_arrays: List of xarray DataArrays from preprocessing",
                "        ",
                "    Returns:",
                "        str: Validation signature symbol",
                "        ",
                "    Raises:",
                "        ValueError: If validation fails",
                "    \"\"\"",
                "    import numpy as np",
                "    ",
                "    if not data_arrays:",
                "        raise ValueError(\"No data arrays provided from preprocessing\")",
                "    ",
                "    reference_shape = None",
                "    ",
                "    for i, data_array in enumerate(data_arrays):",
                "        # Check if it's an xarray DataArray",
                "        if not hasattr(data_array, 'dims') or not hasattr(data_array, 'coords'):",
                "            raise ValueError(f\"Channel {i+1} is not a valid xarray DataArray\")",
                "        ",
                "        # Check coordinates",
                "        if 'time' not in data_array.coords:",
                "            raise ValueError(f\"Channel {i+1} missing time coordinate\")",
                "        ",
                "        spatial_dims = [dim for dim in data_array.dims if dim != 'time']",
                "        if len(spatial_dims) != 2:",
                "            raise ValueError(f\"Channel {i+1} must have exactly 2 spatial dimensions, got {spatial_dims}\")",
                "        ",
                "        for dim in spatial_dims:",
                "            if dim not in data_array.coords:",
                "                raise ValueError(f\"Channel {i+1} missing coordinate: {dim}\")",
                "        ",
                "        # Check shape consistency",
                "        shape = data_array.shape",
                "        if reference_shape is None:",
                "            reference_shape = shape",
                "        else:",
                "            if shape != reference_shape:",
                "                raise ValueError(f\"Channel {i+1} shape {shape} doesn't match reference {reference_shape}\")",
                "    ",
                "    # Generate validation signature",
                "    signature_components = [",
                "        f\"CH{len(data_arrays)}\",  # Channel count",
                "        f\"T{reference_shape[0]}\",  # Time dimension",
                "        f\"S{reference_shape[1]}x{reference_shape[2]}\",  # Spatial dimensions",
                "        f\"DT{data_arrays[0].values.dtype}\",  # Data type",
                "    ]",
                "    ",
                "    signature = \"★PRE_\" + \"_\".join(signature_components) + \"★\"",
                "    ",
                "    return signature",
                "",
            ])
        
        if postprocessing_section:
            script_lines.extend([
                "def validate_postprocessing_output(result_array):",
                "    \"\"\"",
                "    Validate postprocessing output coordinates and data type.",
                "    ",
                "    Args:",
                "        result_array: xarray DataArray from postprocessing",
                "        ",
                "    Returns:",
                "        str: Validation signature symbol",
                "        ",
                "    Raises:",
                "        ValueError: If validation fails",
                "    \"\"\"",
                "    import numpy as np",
                "    ",
                "    # Check if it's an xarray DataArray",
                "    if not hasattr(result_array, 'dims') or not hasattr(result_array, 'coords'):",
                "        raise ValueError(\"Postprocessing output is not a valid xarray DataArray\")",
                "    ",
                "    # Check required coordinates",
                "    if 'time' not in result_array.coords:",
                "        raise ValueError(\"Missing time coordinate\")",
                "    ",
                "    spatial_dims = [dim for dim in result_array.dims if dim != 'time']",
                "    if len(spatial_dims) != 2:",
                "        raise ValueError(f\"Expected 2 spatial dimensions, got {len(spatial_dims)}: {spatial_dims}\")",
                "    ",
                "    for dim in spatial_dims:",
                "        if dim not in result_array.coords:",
                "            raise ValueError(f\"Missing spatial coordinate: {dim}\")",
                "    ",
                "    # Check shape",
                "    shape = result_array.shape",
                "    ",
                "    # Generate validation signature",
                "    signature_components = [",
                "        f\"T{shape[0]}\",  # Time dimension",
                "        f\"S{shape[1]}x{shape[2]}\",  # Spatial dimensions",
                "        f\"DT{result_array.values.dtype}\",  # Data type",
                "    ]",
                "    ",
                "    signature = \"★POST_\" + \"_\".join(signature_components) + \"★\"",
                "    ",
                "    return signature",
                "",
            ])
        
        if preprocessing_section:
            script_lines.extend([
                "def preprocessing(array: Tuple[xr.DataArray, ...]) -> Tuple[xr.DataArray, ...]:",
                preprocessing_section,
                "",
            ])
        
        if postprocessing_section:
            script_lines.extend([
                "def postprocessing(array: xr.DataArray) -> xr.DataArray:",
                postprocessing_section,
                "",
            ])
        
        script_lines.extend([
            "def get_model():",
            f"    logging.info(\"Loading Random Forest model for {virtual_dataset_name}...\")",
            "",
            "    client = storage.Client()",
            f"    bucket = client.get_bucket('{bucket_name}')",
            f"    blob = bucket.blob('{uid}/virtual_datasets/{virtual_dataset_name}/{virtual_dataset_name}.onnx')",
            "",
            "    model = BytesIO()",
            "    blob.download_to_file(model)",
            "    model.seek(0)",
            "",
            "    session = InferenceSession(model.read(), providers=[\"CPUExecutionProvider\"])",
            "    return session",
            "",
            f"def {virtual_product_name}(*bands, model):",
            "    logging.info(\"Start preparing Random Forest data\")",
            "    data_arrays = list(bands)",
            "    ",
            "    if not data_arrays:",
            "        raise ValueError(\"No bands provided\")",
            "    ",
        ])
        
        if preprocessing_section:
            script_lines.extend([
                "    # Apply preprocessing",
                "    data_arrays = preprocessing(tuple(data_arrays))",
                "    data_arrays = list(data_arrays)  # Convert back to list for processing",
                "    ",
                "    # Validate preprocessing output",
                "    preprocessing_signature = validate_preprocessing_output(data_arrays)",
                "    ",
            ])
        
        script_lines.extend([
            "    reference_array = data_arrays[0]",
            "    original_shape = reference_array.shape",
            "    ",
            "    if 'time' in reference_array.dims:",
            "        time_coords = reference_array.coords['time']",
            "        if len(time_coords) == 1:",
            "            output_timestamp = time_coords[0]",
            "        else:",
            "            years = [pd.to_datetime(t).year for t in time_coords.values]",
            "            unique_years = set(years)",
            "            ",
            "            if len(unique_years) == 1:",
            "                year = list(unique_years)[0]",
            "                output_timestamp = pd.Timestamp(f\"{year}-01-01\")",
            "            else:",
            "                latest_year = max(unique_years)",
            "                output_timestamp = pd.Timestamp(f\"{latest_year}-01-01\")",
            "    else:",
            "        output_timestamp = pd.Timestamp(\"1970-01-01\")",
            "",
            "    averaged_bands = []",
            "    for data_array in data_arrays:",
            "        if 'time' in data_array.dims:",
            "            averaged_band = np.mean(data_array.values, axis=0)",
            "        else:",
            "            averaged_band = data_array.values",
            "",
            "        flattened_band = averaged_band.reshape(-1, 1)",
            "        averaged_bands.append(flattened_band)",
            "",
            "    input_data = np.hstack(averaged_bands)",
            "",
            "    output = model.run(None, {\"float_input\": input_data.astype(np.float32)})[0]",
            "",
            "    if len(original_shape) >= 3:",
            "        spatial_shape = original_shape[1:]",
            "    else:",
            "        spatial_shape = original_shape",
            "",
            "    output_reshaped = output.reshape(spatial_shape)",
            "",
            "    output_with_time = np.expand_dims(output_reshaped, axis=0)",
            "",
            "    if 'time' in reference_array.dims:",
            "        spatial_dims = [dim for dim in reference_array.dims if dim != 'time']",
            "        spatial_coords = {dim: reference_array.coords[dim] for dim in spatial_dims if dim in reference_array.coords}",
            "    else:",
            "        spatial_dims = list(reference_array.dims)",
            "        spatial_coords = dict(reference_array.coords)",
            "",
            "    result = xr.DataArray(",
            "        data=output_with_time.astype(np.float32),",
            "        dims=['time'] + list(spatial_dims),",
            "        coords={",
            "            'time': [output_timestamp.values],",
            "            'y': spatial_coords['y'].values,",
            "            'x': spatial_coords['x'].values",
            "        },",
            "        attrs={",
            "            'description': 'Random Forest model prediction',",
            "        }",
            "    )",
        ])
        
        if postprocessing_section:
            script_lines.extend([
                "    # Apply postprocessing",
                "    result = postprocessing(result)",
                "    ",
                "    # Validate postprocessing output",
                "    postprocessing_signature = validate_postprocessing_output(result)",
                "    ",
            ])
        
        script_lines.append("    return result")
        
        return "\n".join(script_lines)

    @require_api_key
    async def _generate_cnn_script(self, bucket_name: str, virtual_dataset_name: str, virtual_product_name: str, processing_script_path: Optional[str] = None) -> str:
        """
        Generate Python inference script for CNN model with time-stacked bands.
        
        Args:
            bucket_name: Name of the bucket where the model is stored
            virtual_dataset_name: Name of the virtual dataset and the model
            virtual_product_name: Name of the virtual product
            processing_script_path: Path to the processing script, if not provided, no processing will be done
        Returns:
            str: Generated Python script content
        """
        user_info = await self._client.auth.get_user_info()
        uid = user_info["uid"]
        preprocessing_code, postprocessing_code = None, None

        if processing_script_path:
            try:
                preprocessing_code, postprocessing_code = self._parse_processing_script(processing_script_path)
                if preprocessing_code:
                    self._client.logger.info(f"Using custom preprocessing from: {processing_script_path}")
                if postprocessing_code:
                    self._client.logger.info(f"Using custom postprocessing from: {processing_script_path}")
                if not preprocessing_code and not postprocessing_code:
                    self._client.logger.warning(f"No preprocessing or postprocessing functions found in {processing_script_path}")
                    self._client.logger.info("Deployment will continue without custom processing")
            except Exception as e:
                raise ValueError(f"Failed to load processing script: {str(e)}")
            
        preprocessing_section = ""
        if preprocessing_code and preprocessing_code.strip():
            clean_preprocessing = textwrap.dedent(preprocessing_code)
            preprocessing_section = textwrap.indent(clean_preprocessing, '    ')
        
        postprocessing_section = ""
        if postprocessing_code and postprocessing_code.strip():
            clean_postprocessing = textwrap.dedent(postprocessing_code)
            postprocessing_section = textwrap.indent(clean_postprocessing, '    ')

        script_lines = [
            "import logging",
            "from io import BytesIO",
            "import numpy as np",
            "import pandas as pd",
            "import xarray as xr",
            "from google.cloud import storage",
            "from onnxruntime import InferenceSession",
            "from typing import Tuple",
            "",
            "logging.basicConfig(",
            "    level=logging.INFO",
            ")",
            "",
        ]
        
        if preprocessing_section:
            script_lines.extend([
                "def validate_preprocessing_output(data_arrays):",
                "    \"\"\"",
                "    Validate preprocessing output coordinates and data type.",
                "    ",
                "    Args:",
                "        data_arrays: List of xarray DataArrays from preprocessing",
                "        ",
                "    Returns:",
                "        str: Validation signature symbol",
                "        ",
                "    Raises:",
                "        ValueError: If validation fails",
                "    \"\"\"",
                "    import numpy as np",
                "    ",
                "    if not data_arrays:",
                "        raise ValueError(\"No data arrays provided from preprocessing\")",
                "    ",
                "    reference_shape = None",
                "    ",
                "    for i, data_array in enumerate(data_arrays):",
                "        # Check if it's an xarray DataArray",
                "        if not hasattr(data_array, 'dims') or not hasattr(data_array, 'coords'):",
                "            raise ValueError(f\"Channel {i+1} is not a valid xarray DataArray\")",
                "        ",
                "        # Check coordinates",
                "        if 'time' not in data_array.coords:",
                "            raise ValueError(f\"Channel {i+1} missing time coordinate\")",
                "        ",
                "        spatial_dims = [dim for dim in data_array.dims if dim != 'time']",
                "        if len(spatial_dims) != 2:",
                "            raise ValueError(f\"Channel {i+1} must have exactly 2 spatial dimensions, got {spatial_dims}\")",
                "        ",
                "        for dim in spatial_dims:",
                "            if dim not in data_array.coords:",
                "                raise ValueError(f\"Channel {i+1} missing coordinate: {dim}\")",
                "        ",
                "        # Check shape consistency",
                "        shape = data_array.shape",
                "        if reference_shape is None:",
                "            reference_shape = shape",
                "        else:",
                "            if shape != reference_shape:",
                "                raise ValueError(f\"Channel {i+1} shape {shape} doesn't match reference {reference_shape}\")",
                "    ",
                "    # Generate validation signature",
                "    signature_components = [",
                "        f\"CH{len(data_arrays)}\",  # Channel count",
                "        f\"T{reference_shape[0]}\",  # Time dimension",
                "        f\"S{reference_shape[1]}x{reference_shape[2]}\",  # Spatial dimensions",
                "        f\"DT{data_arrays[0].values.dtype}\",  # Data type",
                "    ]",
                "    ",
                "    signature = \"★PRE_\" + \"_\".join(signature_components) + \"★\"",
                "    ",
                "    return signature",
                "",
            ])
        
        if postprocessing_section:
            script_lines.extend([
                "def validate_postprocessing_output(result_array):",
                "    \"\"\"",
                "    Validate postprocessing output coordinates and data type.",
                "    ",
                "    Args:",
                "        result_array: xarray DataArray from postprocessing",
                "        ",
                "    Returns:",
                "        str: Validation signature symbol",
                "        ",
                "    Raises:",
                "        ValueError: If validation fails",
                "    \"\"\"",
                "    import numpy as np",
                "    ",
                "    # Check if it's an xarray DataArray",
                "    if not hasattr(result_array, 'dims') or not hasattr(result_array, 'coords'):",
                "        raise ValueError(\"Postprocessing output is not a valid xarray DataArray\")",
                "    ",
                "    # Check required coordinates",
                "    if 'time' not in result_array.coords:",
                "        raise ValueError(\"Missing time coordinate\")",
                "    ",
                "    spatial_dims = [dim for dim in result_array.dims if dim != 'time']",
                "    if len(spatial_dims) != 2:",
                "        raise ValueError(f\"Expected 2 spatial dimensions, got {len(spatial_dims)}: {spatial_dims}\")",
                "    ",
                "    for dim in spatial_dims:",
                "        if dim not in result_array.coords:",
                "            raise ValueError(f\"Missing spatial coordinate: {dim}\")",
                "    ",
                "    # Check shape",
                "    shape = result_array.shape",
                "    ",
                "    # Generate validation signature",
                "    signature_components = [",
                "        f\"T{shape[0]}\",  # Time dimension",
                "        f\"S{shape[1]}x{shape[2]}\",  # Spatial dimensions",
                "        f\"DT{result_array.values.dtype}\",  # Data type",
                "    ]",
                "    ",
                "    signature = \"★POST_\" + \"_\".join(signature_components) + \"★\"",
                "    ",
                "    return signature",
                "",
            ])
        
        if preprocessing_section:
            script_lines.extend([
                "def preprocessing(array: Tuple[xr.DataArray, ...]) -> Tuple[xr.DataArray, ...]:",
                preprocessing_section,
                "",
            ])
        
        if postprocessing_section:
            script_lines.extend([
                "def postprocessing(array: xr.DataArray) -> xr.DataArray:",
                postprocessing_section,
                "",
            ])
        
        script_lines.extend([
            "def get_model():",
            f"    logging.info(\"Loading CNN model for {virtual_dataset_name}...\")",
            "",
            "    client = storage.Client()",
            f"    bucket = client.get_bucket('{bucket_name}')",
            f"    blob = bucket.blob('{uid}/virtual_datasets/{virtual_dataset_name}/{virtual_dataset_name}.onnx')",
            "",
            "    model = BytesIO()",
            "    blob.download_to_file(model)",
            "    model.seek(0)",
            "",
            "    session = InferenceSession(model.read(), providers=[\"CPUExecutionProvider\"])",
            "    return session",
            "",
            f"def {virtual_product_name}(*bands, model):",
            "    logging.info(\"Start preparing CNN data with time-stacked bands\")",
            "    data_arrays = list(bands)",
            "    ",
            "    if not data_arrays:",
            "        raise ValueError(\"No bands provided\")",
            "    ",
        ])
        
        if preprocessing_section:
            script_lines.extend([
                "    # Apply preprocessing",
                "    data_arrays = preprocessing(tuple(data_arrays))",
                "    data_arrays = list(data_arrays)  # Convert back to list for processing",
                "    ",
                "    # Validate preprocessing output",
                "    preprocessing_signature = validate_preprocessing_output(data_arrays)",
                "    ",
            ])
        
        script_lines.extend([
            "    reference_array = data_arrays[0]",
            "    original_shape = reference_array.shape",
            "    ",
            "    # Get time coordinates - all bands should have the same time dimension",
            "    if 'time' not in reference_array.dims:",
            "        raise ValueError(\"Time dimension is required for CNN processing\")",
            "    ",
            "    time_coords = reference_array.coords['time']",
            "    num_timestamps = len(time_coords)",
            "    ",
            "    # Get spatial dimensions",
            "    spatial_dims = [dim for dim in reference_array.dims if dim != 'time']",
            "    height = reference_array.sizes[spatial_dims[0]]  # assuming first spatial dim is height",
            "    width = reference_array.sizes[spatial_dims[1]]   # assuming second spatial dim is width",
            "    ",
            "    # Stack bands across time dimension",
            "    # Result will be: (num_bands * num_timestamps, height, width)",
            "    stacked_channels = []",
            "    ",
            "    for band_idx, data_array in enumerate(data_arrays):",
            "        # Ensure consistent time coordinates across bands",
            "        if not np.array_equal(data_array.coords['time'].values, time_coords.values):",
            "            data_array = data_array.sel(time=time_coords, method='nearest')",
            "        ",
            "        # Extract values and ensure proper ordering (time, height, width)",
            "        band_values = data_array.values",
            "        if band_values.ndim == 3:",
            "            # Reorder dimensions if needed to ensure (time, height, width)",
            "            time_dim_idx = data_array.dims.index('time')",
            "            if time_dim_idx != 0:",
            "                axes_order = [time_dim_idx] + [i for i in range(len(data_array.dims)) if i != time_dim_idx]",
            "                band_values = np.transpose(band_values, axes_order)",
            "        ",
            "        # Add each timestamp of this band to the channel stack",
            "        for t in range(num_timestamps):",
            "            stacked_channels.append(band_values[t])",
            "    ",
            "    # Stack all channels: (num_bands * num_timestamps, height, width)",
            "    input_channels = np.stack(stacked_channels, axis=0)",
            "    total_channels = len(data_arrays) * num_timestamps",
            "    ",
            "    # Add batch dimension: (1, num_channels, height, width)",
            "    input_data = np.expand_dims(input_channels, axis=0).astype(np.float32)",
            "    ",
            "    # Run inference",
            "    output = model.run(None, {\"float_input\": input_data})[0]",
            "    ",
            "    # Handle multi-class CNN output properly",
            "    if output.ndim == 4:",
            "        if output.shape[1] == 1:",
            "            # Single class output (regression or binary classification)",
            "            output_2d = output[0, 0]",
            "        else:",
            "            # Multi-class output - convert logits/probabilities to class predictions",
            "            output_classes = np.argmax(output, axis=1)  # Shape: (1, height, width)",
            "            output_2d = output_classes[0]  # Shape: (height, width)",
            "            ",
            "            # Apply class merging: merge class 6 into class 3",
            "            output_2d = np.where(output_2d == 6, 3, output_2d)",
            "    elif output.ndim == 3:",
            "        # Remove batch dimension",
            "        output_2d = output[0]",
            "    else:",
            "        # Handle other cases",
            "        output_2d = np.squeeze(output)",
            "        if output_2d.ndim != 2:",
            "            raise ValueError(f\"Unexpected output shape after processing: {output_2d.shape}\")",
            "    ",
            "    # Ensure output is 2D",
            "    if output_2d.ndim != 2:",
            "        raise ValueError(f\"Final output must be 2D, got shape: {output_2d.shape}\")",
            "    ",
            "    # Determine output timestamp (use the latest timestamp)",
            "    output_timestamp = time_coords[-1]",
            "    ",
            "    # Get spatial coordinates from reference array",
            "    spatial_coords = {dim: reference_array.coords[dim] for dim in spatial_dims}",
            "    ",
            "    # Create output DataArray with appropriate data type",
            "    # Use int32 for classification, float32 for regression",
            "    is_multiclass = output.ndim == 4 and output.shape[1] > 1",
            "    if is_multiclass:",
            "        # Multi-class classification - use integer type",
            "        output_dtype = np.int32",
            "    else:",
            "        # Single output - use float type",
            "        output_dtype = np.float32",
            "    ",
            "    result = xr.DataArray(",
            "        data=np.expand_dims(output_2d.astype(output_dtype), axis=0),",
            "        dims=['time'] + spatial_dims,",
            "        coords={",
            "            'time': [output_timestamp.values],",
            "            spatial_dims[0]: spatial_coords[spatial_dims[0]].values,",
            "            spatial_dims[1]: spatial_coords[spatial_dims[1]].values",
            "        },",
            "        attrs={",
            "            'description': 'CNN model prediction',",
            "        }",
            "    )",
        ])
        
        if postprocessing_section:
            script_lines.extend([
                "    # Apply postprocessing",
                "    result = postprocessing(result)",
                "    ",
                "    # Validate postprocessing output",
                "    postprocessing_signature = validate_postprocessing_output(result)",
                "    ",
            ])
        
        script_lines.append("    return result")
        
        return "\n".join(script_lines)

    def _parse_processing_script(self, script_path: str) -> Tuple[Optional[str], Optional[str]]:
        """
        Parse a Python file and extract preprocessing and postprocessing function bodies.
        
        Args:
            script_path: Path to the Python file containing processing functions
            
        Returns:
            Tuple of (preprocessing_code, postprocessing_code) where each can be None
        """
        try:
            with open(script_path, 'r', encoding='utf-8') as f:
                script_content = f.read()
        except FileNotFoundError:
            raise FileNotFoundError(f"Processing script not found: {script_path}")
        except Exception as e:
            raise ValueError(f"Error reading processing script: {e}")
        
        if not script_content.strip():
            self._client.logger.info(f"Processing script {script_path} is empty")
            return None, None
        
        try:
            tree = ast.parse(script_content)
        except SyntaxError as e:
            raise ValueError(f"Syntax error in processing script: {e}")
        
        preprocessing_code = None
        postprocessing_code = None
        
        function_names = []
        for node in ast.walk(tree):
            if isinstance(node, ast.FunctionDef):
                function_names.append(node.name)
                if node.name == 'preprocessing':
                    preprocessing_code = self._extract_function_body(script_content, node)
                elif node.name == 'postprocessing':
                    postprocessing_code = self._extract_function_body(script_content, node)
        
        if not function_names:
            self._client.logger.warning(f"No functions found in processing script: {script_path}")
        else:
            found_functions = [name for name in function_names if name in ['preprocessing', 'postprocessing']]
            if found_functions:
                self._client.logger.info(f"Found processing functions: {found_functions}")
            else:
                self._client.logger.warning(f"No 'preprocessing' or 'postprocessing' functions found in {script_path}. "
                                          f"Available functions: {function_names}")
        
        return preprocessing_code, postprocessing_code

    def _extract_function_body(self, script_content: str, func_node: ast.FunctionDef) -> str:
        """Extract the body of a function from the script content."""
        lines = script_content.split('\n')
        
        start_line = func_node.lineno - 1
        end_line = func_node.end_lineno - 1 if hasattr(func_node, 'end_lineno') else len(lines) - 1
        
        body_lines = []
        for i in range(start_line + 1, end_line + 1):
            if i < len(lines):
                body_lines.append(lines[i])
        
        if not body_lines:
            return ""
        
        body_text = '\n'.join(body_lines)
        cleaned_body = textwrap.dedent(body_text).strip()
        
        if not cleaned_body or cleaned_body in ['pass', 'return', 'return None']:
            return ""
        
        return cleaned_body
    
    def _convert_model_to_onnx(self, model, input_shape: Tuple[int, ...] = None, model_type: Optional[str] = None) -> bytes:
        """
        Convert a model to ONNX format and return as bytes.
        
        Args:
            model: The model object (PyTorch or scikit-learn)
            input_shape: Shape of input data
            model_type: Type of model (neural_network, random_forest), only used for onnx model generation
        Returns:
            bytes: ONNX model as bytes
            
        Raises:
            ValueError: If model type is not supported
            ImportError: If required libraries are not installed
        """
        if isinstance(model, torch.nn.Module):
            if not TORCH_AVAILABLE:
                raise ImportError("PyTorch is not installed. Please install it with: pip install torch")
            return self._convert_pytorch_to_onnx(model, input_shape), "neural_network"
        elif isinstance(model, BaseEstimator):
            if not SKL2ONNX_AVAILABLE:
                raise ImportError("skl2onnx is not installed. Please install it with: pip install skl2onnx")
            return self._convert_sklearn_to_onnx(model, input_shape), "random_forest"
        elif isinstance(model, ort.InferenceSession):
            if model_type is None:
                raise ValueError(
                    "For ONNX InferenceSession models, you must specify the 'model_type' parameter. Currently 'nerual network' and 'random forest' are supported."
                    "Example: model_type='random forest' or model_type='neural network'"
                )
            return model.SerializeToString(), model_type
        else:
            model_type = type(model).__name__
            raise ValueError(f"Unsupported model type: {model_type}. Supported types: PyTorch nn.Module, sklearn BaseEstimator")

    def _convert_pytorch_to_onnx(self, model, input_shape: Tuple[int, ...]) -> bytes:
        try:
            model.eval()
            dummy_input = torch.randn(input_shape)
            onnx_buffer = BytesIO()
            
            if len(input_shape) == 4:
                dynamic_axes = {
                    'float_input': {
                        0: 'batch_size',
                        2: 'height',
                        3: 'width'
                    }
                }
                
            elif len(input_shape) == 5:
                dynamic_axes = {
                    'float_input': {
                        0: 'batch_size',
                        3: 'height',
                        4: 'width'
                    }
                }
                
            else:
                dynamic_axes = {
                    'float_input': {
                        0: 'batch_size'
                    }
                }
            
            torch.onnx.export(
                model,
                dummy_input,
                onnx_buffer,
                input_names=['float_input'],
                dynamic_axes=dynamic_axes
            )
            
            return onnx_buffer.getvalue()
        except Exception as e:
            raise ValueError(f"Failed to convert PyTorch model to ONNX: {str(e)}")

    def _convert_sklearn_to_onnx(self, model, input_shape: Tuple[int, ...]) -> bytes:
        """
        Convert scikit-learn model(assume it is a random forest model) to ONNX format.
        
        Args:
            model: The scikit-learn model object
            input_shape: Shape of input data (required)
            
        Returns:
            bytes: ONNX model as bytes
            
        Raises:
            ValueError: If conversion fails
        """
        self._client.logger.info(f"Converting random forest model to ONNX...")
        
        try:
            initial_type = [('float_input', FloatTensorType(input_shape))]
            onnx_model = convert_sklearn(model, initial_types=initial_type)
            return onnx_model.SerializeToString()
        except Exception as e:
            raise ValueError(f"Failed to convert scikit-learn model to ONNX: {str(e)}")

    @require_api_key
    def train_model(
        self, 
        model_name: str, 
        training_dataset: str, 
        task_type: str, 
        model_category: str, 
        architecture: str, 
        region: str, 
        hyperparameters: dict = None
    ) -> dict:
        """
        Train a model using the external model training API.
        
        Args:
            model_name (str): The name of the model to train.
            training_dataset (str): The training dataset identifier.
            task_type (str): The type of ML task (e.g., regression, classification).
            model_category (str): The category of model (e.g., random_forest).
            architecture (str): The model architecture.
            region (str): The region identifier.
            hyperparameters (dict, optional): Additional hyperparameters for training.
            
        Returns:
            dict: The response from the model training API.
            
        Raises:
            APIError: If the API request fails
        """
        payload = {
            "model_name": model_name,
            "training_dataset": training_dataset,
            "task_type": task_type,
            "model_category": model_category,
            "architecture": architecture,
            "region": region,
            "hyperparameters": hyperparameters
        }
        return self._client._terrakio_request("POST", "/train_model", json=payload)



import os
import asyncio
import tempfile
import time
import pandas as pd
import geopandas as gpd
from geopandas import GeoDataFrame
from shapely.geometry import mapping
from pathlib import Path
from ..exceptions import APIError, ConfigurationError
from ..helper.bounded_taskgroup import BoundedTaskGroup
from ..helper.tiles import tiles
import uuid
import xarray as xr
import random
import psutil
import copy
from shapely.geometry import shape
from shapely.ops import transform
from shapely.geometry import box
import pyproj

import pandas as pd
import geopandas as gpd

from typing import Optional

def expand_on_time(gdf):
    """
    Expand datasets on time dimension - each time becomes a new row.
    
    Input: GeoDataFrame with 'geometry' and 'dataset' columns (or variable columns)
    Output: GeoDataFrame with time in multi-index and datasets without time coordinate
    """
    rows = []
    
    for idx, row in gdf.iterrows():
        if 'geometry' in gdf.columns:
            geometry = row['geometry']
        elif gdf.index.name == 'geometry':
            geometry = idx
        else:
            raise ValueError(f"Cannot find geometry in columns: {list(gdf.columns)} or index: {gdf.index.name}")
        
        if 'dataset' in gdf.columns:
            dataset = row['dataset']
            
            if 'time' in dataset.dims:
                for time_val in dataset.time.values:
                    time_slice = dataset.sel(time=time_val).drop_vars('time')
                    rows.append({
                        'geometry': geometry,
                        'time': time_val,
                        'dataset': time_slice
                    })
            else:
                rows.append({
                    'geometry': geometry,
                    'dataset': dataset
                })
        else:
            variable_columns = list(gdf.columns)
            
            first_dataset = row[variable_columns[0]]
            if 'time' in first_dataset.dims:
                time_values = first_dataset.time.values
                
                for time_val in time_values:
                    row_data = {'geometry': geometry, 'time': time_val}
                    
                    for var_col in variable_columns:
                        dataset = row[var_col]
                        time_slice = dataset.sel(time=time_val).drop_vars('time')
                        row_data[var_col] = time_slice
                    
                    rows.append(row_data)
            else:
                row_data = {'geometry': geometry}
                for var_col in variable_columns:
                    row_data[var_col] = row[var_col]
                rows.append(row_data)
    
    result_df = pd.DataFrame(rows)
    
    if 'time' in result_df.columns:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry', 'time'])
    else:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry'])
    
    return result_gdf

def expand_on_variables(gdf):
    """
    Expand datasets on variables dimension - each variable becomes a new column.
    
    Input: GeoDataFrame with 'geometry' and 'dataset' columns (or already time-expanded)
    Output: GeoDataFrame with separate column for each variable
    """
    rows = []
    
    for idx, row in gdf.iterrows():
        if 'geometry' in gdf.columns:
            geometry = row['geometry']
        elif hasattr(gdf.index, 'names') and 'geometry' in gdf.index.names:
            if isinstance(idx, tuple):
                geometry_idx = gdf.index.names.index('geometry')
                geometry = idx[geometry_idx]
                time_idx = gdf.index.names.index('time')
                time_val = idx[time_idx]
            else:
                geometry = idx
                time_val = None
        else:
            raise ValueError(f"Cannot find geometry in columns: {list(gdf.columns)} or index: {gdf.index.names}")
        
        if 'dataset' in gdf.columns:
            dataset = row['dataset']
            
            var_names = list(dataset.data_vars.keys())
            
            if len(var_names) <= 1:
                if len(var_names) == 0:
                    continue
            
            if hasattr(gdf.index, 'names') and 'time' in gdf.index.names:
                row_data = {'geometry': geometry, 'time': time_val}
            else:
                row_data = {'geometry': geometry}
            
            for var_name in var_names:
                var_dataset = dataset[[var_name]]
                
                if len(var_dataset.dims) == 0:
                    row_data[var_name] = float(var_dataset[var_name].values)
                else:
                    row_data[var_name] = var_dataset
            
            rows.append(row_data)
        else:
            raise ValueError("Expected 'dataset' column for variable expansion")
    
    result_df = pd.DataFrame(rows)
    
    if 'time' in result_df.columns:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry', 'time'])
    else:
        result_gdf = gpd.GeoDataFrame(result_df, geometry='geometry')
        result_gdf = result_gdf.set_index(['geometry'])
    
    return result_gdf


def expand_on_variables_and_time(gdf):
    """
    Convenience function to expand on both variables and time.
    Automatically detects which expansions are possible.
    """
    try:
        expanded_on_time = expand_on_time(gdf)
    except Exception as e:
        expanded_on_time = gdf
    
    try:
        expanded_on_variables_and_time = expand_on_variables(expanded_on_time)
        return expanded_on_variables_and_time
    except Exception as e:
        return expanded_on_time

def estimate_geometry_size_ratio(queries: list):
    """Calculate size ratios for all geometries relative to the first geometry using bounding box area."""
    
    areas = []
    
    for query in queries:
        geom = shape(query["feature"]["geometry"])
        in_crs = query["in_crs"]
        
        if in_crs and in_crs != 'EPSG:3857':
            transformer = pyproj.Transformer.from_crs(in_crs, 'EPSG:3857', always_xy=True)
            transformed_geom = transform(transformer.transform, geom)
            bbox = box(*transformed_geom.bounds)
            area = bbox.area
        else:
            bbox = box(*geom.bounds)
            area = bbox.area
        
        areas.append(area)    
    base_area = areas[0]
    
    if base_area == 0:
        non_zero_areas = [area for area in areas if area > 0]
        base_area = non_zero_areas[0] if non_zero_areas else 1.0
    
    ratios = []
    for area in areas:
        if area == 0:
            ratios.append(0.1)
        else:
            ratios.append(area / base_area)
    
    return ratios

async def estimate_query_size(
    client,
    quries: list[dict],
):
    first_query = quries[0]

    first_query_dataset = await client.geoquery(**first_query)
    ratios = estimate_geometry_size_ratio(quries)
    total_size_mb = 0
    for i in range(len(ratios)):
        total_size_mb += first_query_dataset.nbytes * ratios[i] / (1024**2)
    return total_size_mb

async def request_geoquery_list(
        client,
        quries: list[dict],
        conc: int = 20,
):
    """
    Execute multiple geo queries.
    
    Args:
        client: The Terrakio client instance
        quries: List of dictionaries containing query parameters
        conc: The concurrency level for the requests
        
    Returns:
        List of query results
        
    Raises:
        ValueError: If the queries list is empty
    """
    if not quries:
        raise ValueError("Queries list cannot be empty")
    if conc > 100:
        raise ValueError("Concurrency (conc) is too high. Please set conc to 100 or less.")

    for i, query in enumerate(quries):
        if 'expr' not in query:
            raise ValueError(f"Query at index {i} is missing the required 'expr' key")
        if 'feature' not in query:
            raise ValueError(f"Query at index {i} is missing the required 'feature' key")
        if 'in_crs' not in query:
            raise ValueError(f"Query at index {i} is missing the required 'in_crs' key")
    
    completed_count = 0
    lock = asyncio.Lock()
    async def single_geo_query(query):
        """
        Execute multiple geo queries concurrently.
        
        Args:
            quries: List of dictionaries containing query parameters
        """
        total_number_of_requests = len(quries)
        nonlocal completed_count
        try:
            result = await client.geoquery(**query)
            if isinstance(result, dict) and result.get("error"):
                error_msg = f"Request failed: {result.get('error_message', 'Unknown error')}"
                if result.get('status_code'):
                    error_msg = f"Request failed with status {result['status_code']}: {result.get('error_message', 'Unknown error')}"
                raise APIError(error_msg)
            if isinstance(result, list):
                result = result[0]
                timestamp_number = result['request_count']
                return timestamp_number
            if not isinstance(result, xr.Dataset):
                raise ValueError(f"Expected xarray Dataset, got {type(result)}")
            
            async with lock:
                completed_count += 1
                if completed_count % max(1, total_number_of_requests // 10) == 0:
                    client.logger.info(f"Progress: {completed_count}/{total_number_of_requests} requests processed")
            return result   
        except Exception as e:
            async with lock:
                completed_count += 1
            raise
    
    try:
        async with BoundedTaskGroup(max_concurrency=conc) as tg:
            tasks = [tg.create_task(single_geo_query(quries[idx])) for idx in range(len(quries))]
        all_results = [task.result() for task in tasks]

    except* Exception as eg:
        for e in eg.exceptions:
            if hasattr(e, 'response'):
                raise APIError(f"API request failed: {e.response.text}")
        raise
    client.logger.info("All requests completed!")
     
    if not all_results:
        raise ValueError("No valid results were returned for any geometry")
    if isinstance(all_results, list) and type(all_results[0]) == int:
        return sum(all_results)/len(all_results)
    else:
        geometries = []
        for query in quries:
            feature = query['feature']
            geometry = shape(feature['geometry'])
            geometries.append(geometry)
        result_gdf = gpd.GeoDataFrame({
            'geometry': geometries,
            'dataset': all_results
        })
        return result_gdf

async def estimate_timestamp_number(
        client,
        quries: list[dict],
):
    if len(quries) <= 3:
        return quries
    sampled_queries = [query.copy() for query in random.sample(quries, 3)]
    for query in sampled_queries:
        query['debug'] = 'grpc'
    result = await request_geoquery_list(client = client, quries = sampled_queries, conc = 5)
    total_estimated_number_of_timestamps = result * len(quries)
    return total_estimated_number_of_timestamps


def get_available_memory_mb():
    """
    Get available system memory in MB
    
    Returns:
        float: Available memory in MB
    """
    memory = psutil.virtual_memory()
    available_mb = memory.available / (1024 * 1024)
    return round(available_mb, 2)

async def local_or_remote(
        client,
        quries: list[dict],
):
    if len(quries) > 1000:
        return {
            "local_or_remote": "remote",
            "reason": "The number of the requests is too large(>1000), please set the mass_stats parameter to True",
        }
    elif await estimate_timestamp_number(client = client, quries = quries) > 25000:
        return {
            "local_or_remote": "remote",
            "reason": "The time taking for making these requests is too long, please set the mass_stats parameter to True",
        }
    elif await estimate_query_size(client = client, quries = quries) > get_available_memory_mb():
        return {
            "local_or_remote": "remote",
            "reason": "The size of the dataset is too large, please set the mass_stats parameter to True",
        }
    else:
        return {
            "local_or_remote": "local",
            "reason": "The number of the requests is not too large, and the time taking for making these requests is not too long, and the size of the dataset is not too large",
        }
    
def gdf_to_json(
    gdf: GeoDataFrame,
    expr: str,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    id_column: Optional[str] = None,
):
    """
    Convert a GeoDataFrame to a list of JSON requests for mass_stats processing.
    
    Args:
        gdf: GeoDataFrame containing geometries and optional metadata
        expr: Expression to evaluate
        in_crs: Input coordinate reference system
        out_crs: Output coordinate reference system
        resolution: Resolution parameter
        geom_fix: Whether to fix geometry issues
        id_column: Optional column name to use for group and file names
        
    Returns:
        list: List of dictionaries formatted for mass_stats requests
    """
    mass_stats_requests = []
    
    # Loop through each row in the GeoDataFrame
    for idx, row in gdf.iterrows():
        # Create the request feature
        request_feature = {
            "expr": expr,
            "feature": {
                "type": "Feature",
                "geometry": mapping(gdf.geometry.iloc[idx]),
                "properties": {}
            },
            "in_crs": in_crs,
            "out_crs": out_crs,
            "resolution": resolution,
            "geom_fix": geom_fix,
        }
        
        # Determine group name and file name based on id_column
        if id_column is not None and id_column in gdf.columns:
            # Use the value from the specified column as group and file name
            identifier = str(row[id_column])
            group_name = f"group_{identifier}"
            file_name = f"file_{identifier}"
        else:
            # Use the index as group and file name
            group_name = f"group_{idx}"
            file_name = f"file_{idx}"
            
        # Create the complete request entry
        request_entry = {
            "group": group_name,
            "file": file_name,
            "request": request_feature,
        }
        
        # Add the request to our list
        mass_stats_requests.append(request_entry)
        
    return mass_stats_requests
    
async def handle_mass_stats(
    client,
    gdf: GeoDataFrame,
    expr: str,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    id_column: Optional[str] = None,

):
    # we have the handle mass stats function, we need to have the list of quries, and we need to pass the quries to the mass stats function
    # we have the three different variables
    
    # Check if id_column is provided
    # if id_column is None:
        # Handle case where no ID column is specified
        # this means that the id column is none, so we could just use the default value of 1 2 3 4
    request_json = gdf_to_json(gdf = gdf, expr = expr, in_crs = in_crs, out_crs = out_crs, resolution = resolution, geom_fix = geom_fix, id_column = id_column)
    # we need to call the execute job function
    job_id =await client.mass_stats.execute_job(
        name = "zonal_stats_job",
        output = "netcdf",
        config = {},
        request_json = request_json,
        overwrite = True,
    )
    return job_id
# async def test_regular_async_mass_stats(regular_async_client):
#     """Test mass statistics with regular client async"""
#     start_result = await regular_async_client.mass_stats.execute_job(
#         name="test_regular_mass_stats_test",
#         region="aus",
#         output="csv",
#         config={},
#         request_json = "./test_config.json",
#         manifest_json = "./test_manifest.json",
#         overwrite=True,
#     )
#     assert isinstance(start_result, dict)
#     assert 'task_id' in start_result
        
        # return 
    # else:
    #     # Handle case where ID column is specified
    #     # Verify the column exists in the GeoDataFrame
        
    #     if id_column not in gdf.columns:
    #         raise ValueError(f"ID column '{id_column}' not found in GeoDataFrame columns: {list(gdf.columns)}")
    # pass
    # the second case is that we have an id_column, we need to use the id_column to create the group name

# we have the mass stats as one of the parameters, so that when a user wants a mass 
# for both cases we need to have the list of quries
async def zonal_stats(
    client,
    gdf: GeoDataFrame,
    expr: str,
    conc: int = 20,
    in_crs: str = "epsg:4326",
    out_crs: str = "epsg:4326",
    resolution: int = -1,
    geom_fix: bool = False,
    mass_stats: bool = False,
    id_column: Optional[str] = None,
):
    """Compute zonal statistics for all geometries in a GeoDataFrame."""

    if mass_stats:
        mass_stats_id = await handle_mass_stats(
            client = client,
            gdf = gdf,
            expr = expr,
            in_crs = in_crs,
            out_crs = out_crs,
            resolution = resolution,
            geom_fix = geom_fix,
            id_column = id_column
        )
        # if we started the mass stats job, we need to return the job id 
        return mass_stats_id
    quries = []
    for i in range(len(gdf)):
        quries.append({
            "expr": expr,
            "feature": {
                "type": "Feature",
                "geometry": mapping(gdf.geometry.iloc[i]),
                "properties": {}
            },
            "in_crs": in_crs,
            "out_crs": out_crs,
            "resolution": resolution,
            "geom_fix": geom_fix,
        })

    local_or_remote_result = await local_or_remote(client= client, quries = quries)
    if local_or_remote_result["local_or_remote"] == "remote":
        raise ValueError(local_or_remote_result["reason"])
    else:
        gdf_with_datasets = await request_geoquery_list(client = client, quries = quries, conc = conc)
        gdf_with_datasets = expand_on_variables_and_time(gdf_with_datasets)
    return gdf_with_datasets

async def create_dataset_file(
    client,
    aoi: str,
    expression: str,
    output: str,
    in_crs: str = "epsg:4326",
    res: float = 0.0001,
    region: str = "aus",
    to_crs: str = "epsg:4326",
    overwrite: bool = True,
    skip_existing: bool = False,
    non_interactive: bool = True,
    poll_interval: int = 30,
    download_path: str = "/home/user/Downloads",
) -> dict:
    
    name = f"tiles-{uuid.uuid4().hex[:8]}"
    
    body, reqs, groups = tiles(
        name = name, 
        aoi = aoi, 
        expression = expression,
        output = output,
        tile_size = 128,
        crs = in_crs,
        res = res,
        region = region,
        to_crs = to_crs,
        fully_cover = True,
        overwrite = overwrite,
        skip_existing = skip_existing,
        non_interactive = non_interactive
    )
    with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as tempreq:
        tempreq.write(reqs)
        tempreqname = tempreq.name
    with tempfile.NamedTemporaryFile(mode='w', suffix='.json', delete=False) as tempmanifest:
        tempmanifest.write(groups)
        tempmanifestname = tempmanifest.name

    task_id = await client.mass_stats.execute_job(
        name=body["name"],
        region=body["region"],
        output=body["output"],
        config = {},
        overwrite=body["overwrite"],
        skip_existing=body["skip_existing"],
        request_json=tempreqname,
        manifest_json=tempmanifestname,
    )

    start_time = time.time()
    status = None
    
    while True:
        try:
            taskid = task_id['task_id']
            trackinfo = await client.mass_stats.track_job([taskid])
            client.logger.info("the trackinfo is: ", trackinfo)
            status = trackinfo[taskid]['status']
            
            if status == 'Completed':
                client.logger.info('Tiles generated successfully!')
                break
            elif status in ['Failed', 'Cancelled', 'Error']:
                raise RuntimeError(f"Job {taskid} failed with status: {status}")
            else:
                elapsed_time = time.time() - start_time
                client.logger.info(f"Job status: {status} - Elapsed time: {elapsed_time:.1f}s", end='\r')
                
                await asyncio.sleep(poll_interval)
                
                
        except KeyboardInterrupt:
            client.logger.info(f"\nInterrupted! Job {taskid} is still running in the background.")
            raise
        except Exception as e:
            client.logger.info(f"\nError tracking job: {e}")
            raise

    os.unlink(tempreqname)
    os.unlink(tempmanifestname)

    combine_result = await client.mass_stats.combine_tiles(body["name"], body["overwrite"], body["output"])
    combine_task_id = combine_result.get("task_id")

    combine_start_time = time.time()
    while True:
        try:
            trackinfo = await client.mass_stats.track_job([combine_task_id])
            client.logger.info('client create dataset file track info:', trackinfo)
            if body["output"] == "netcdf":
                download_file_name = trackinfo[combine_task_id]['folder'] + '.nc'
            elif body["output"] == "geotiff":
                download_file_name = trackinfo[combine_task_id]['folder'] + '.tif'
            bucket = trackinfo[combine_task_id]['bucket']
            combine_status = trackinfo[combine_task_id]['status']
            if combine_status == 'Completed':
                client.logger.info('Tiles combined successfully!')
                break
            elif combine_status in ['Failed', 'Cancelled', 'Error']:
                raise RuntimeError(f"Combine job {combine_task_id} failed with status: {combine_status}")
            else:
                elapsed_time = time.time() - combine_start_time
                client.logger.info(f"Combine job status: {combine_status} - Elapsed time: {elapsed_time:.1f}s", end='\r')
                time.sleep(poll_interval)
        except KeyboardInterrupt:
            client.logger.info(f"\nInterrupted! Combine job {combine_task_id} is still running in the background.")
            raise
        except Exception as e:
            client.logger.info(f"\nError tracking combine job: {e}")
            raise

    if download_path:
        await client.mass_stats.download_file(
            job_name=body["name"],
            bucket=bucket,
            file_type='processed',
            page_size=10,
            output_path=download_path,
        )
    else:
        path = f"{body['name']}/outputs/merged/{download_file_name}"
        client.logger.info(f"Combined file is available at {path}")

    return {"generation_task_id": task_id, "combine_task_id": combine_task_id}



import os
import json
from typing import Dict, Any
from ..exceptions import APIError
from ..helper.decorators import require_token, require_api_key, require_auth

class AuthClient:
    def __init__(self, client):
        self._client = client

    async def signup(self, email: str, password: str) -> Dict[str, str]:
        """
        Signup a new user with email and password.

        Args:
            email: User's email address
            password: User's password
            
        Returns:
            Dict containing the authentication token
            
        Raises:
            APIError: If the signup request fails
        """
        payload = {
            "email": email,
            "password": password
        }

        
        try:
            result = await self._client._terrakio_request("POST", "/users/signup", json=payload)
        except Exception as e:
            self._client.logger.info(f"Signup failed: {str(e)}")
            raise APIError(f"Signup request failed: {str(e)}")
            

    async def login(self, email: str, password: str) -> Dict[str, str]:
        """
        Login a user with email and password.
        
        Args:
            email: User's email address
            password: User's password
            
        Returns:
            Dict containing the authentication token
            
        Raises:
            APIError: If the login request fails
        """
        payload = {
            "email": email,
            "password": password
        }
        
        try:
            result = await self._client._terrakio_request("POST", "/users/login", json=payload)
            token_response = result.get("token")
            
            if token_response:
                self._client.token = token_response

                api_key_response = await self.view_api_key()
                self._client.key = api_key_response
 
                if not self._client.url:
                    self._client.url = "https://api.terrak.io"
                
                self._save_config(email, token_response)
                
                self._client.logger.info(f"Successfully authenticated as: {email}")
                self._client.logger.info(f"Using Terrakio API at: {self._client.url}")
                
            return {"token": token_response} if token_response else {"error": "Login failed"}
            
        except Exception as e:
            self._client.logger.info(f"Login failed: {str(e)}")
            raise APIError(f"Login request failed: {str(e)}")

    @require_token
    async def view_api_key(self) -> str:
        """
        View the current API key for the authenticated user.
        
        Returns:
            str: The API key
            
        Raises:
            APIError: If the API request fails
        """
        result = await self._client._terrakio_request("GET", "/users/key")
        api_key = result.get("apiKey")
        return api_key

    @require_api_key
    @require_token
    async def refresh_api_key(self) -> str:
        """
        Refresh the API key for the authenticated user.
        
        Returns:
            str: The new API key
            
        Raises:
            APIError: If the API request fails
        """
        result = await self._client._terrakio_request("POST", "/users/refresh_key")
        self._client.key = result.get("apiKey")
                
        self._update_config_key()
        
        return self._client.key

    @require_api_key
    def get_user_info(self) -> Dict[str, Any]:
        """
        Get information about the authenticated user.
        
        Returns:
            Dict[str, Any]: User information
            
        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", "/users/info")

    def _save_config(self, email: str, token: str):
        """
        Helper method to save config file.
        
        Args:
            email: User's email address
            token: Authentication token
        """
        config_path = os.path.join(os.environ.get("HOME", ""), ".tkio_config.json")
        
        try:
            config = {"EMAIL": email, "TERRAKIO_API_KEY": self._client.key}
            if os.path.exists(config_path):
                with open(config_path, 'r') as f:
                    config = json.load(f)
                    
            config["EMAIL"] = email
            config["TERRAKIO_API_KEY"] = self._client.key
            config["PERSONAL_TOKEN"] = token
            os.makedirs(os.path.dirname(config_path), exist_ok=True)
            
            with open(config_path, 'w') as f:
                json.dump(config, f, indent=4)
                
            self._client.logger.info(f"API key saved to {config_path}")
            
        except Exception as e:
            self._client.logger.info(f"Warning: Failed to update config file: {e}")

    def _update_config_key(self):
        """
        Helper method to update just the API key in config.
        """
        config_path = os.path.join(os.environ.get("HOME", ""), ".tkio_config.json")
        
        try:
            config = {"EMAIL": "", "TERRAKIO_API_KEY": ""}
            
            if os.path.exists(config_path):
                with open(config_path, 'r') as f:
                    config = json.load(f)
                    
            config["TERRAKIO_API_KEY"] = self._client.key
            
            os.makedirs(os.path.dirname(config_path), exist_ok=True)
            
            with open(config_path, 'w') as f:
                json.dump(config, f, indent=4)
                
            self._client.logger.info(f"API key updated in {config_path}")
            
        except Exception as e:
            self._client.logger.info(f"Warning: Failed to update config file: {e}")


# we have four different circumstances:
# same expression, different region
# same expression, same region
# different expression, same region
# different expression, different region


import pandas as pd
import geopandas as gpd
import xarray as xr
import numpy as np
from typing import Optional, Union, List

@pd.api.extensions.register_dataframe_accessor("geo")
class GeoXarrayAccessor:
    """
    Custom accessor for GeoDataFrames containing xarray datasets or dataarrays.
    Handles both direct xarray objects and lists containing xarray objects.
    Can aggregate across time when time dimension has been expanded into the index.
    """
    
    def __init__(self, pandas_obj):
        self._obj = pandas_obj
        self._validate()
    
    def _validate(self):
        """Validate that the DataFrame has the expected structure."""
        if not isinstance(self._obj, gpd.GeoDataFrame):
            raise AttributeError("Can only use .geo accessor with GeoDataFrames")
        
        # Check for columns with xarray data (including lists containing xarray objects)
        self._xarray_columns = []
        for col in self._obj.columns:
            if col != 'geometry':
                sample_value = self._obj[col].iloc[0] if len(self._obj) > 0 else None
                
                # Check if it's directly an xarray object
                if isinstance(sample_value, (xr.Dataset, xr.DataArray)):
                    self._xarray_columns.append(col)
                # Check if it's a list containing xarray objects
                elif isinstance(sample_value, list) and len(sample_value) > 0:
                    if isinstance(sample_value[0], (xr.Dataset, xr.DataArray)):
                        self._xarray_columns.append(col)
        
        if not self._xarray_columns:
            raise AttributeError("No xarray Dataset or DataArray columns found")
    
    def _extract_xarray_object(self, value):
        """Extract xarray object from various formats (direct object, list, etc.)."""
        if isinstance(value, (xr.Dataset, xr.DataArray)):
            return value
        elif isinstance(value, list) and len(value) > 0:
            if isinstance(value[0], (xr.Dataset, xr.DataArray)):
                return value[0]  # Take the first item from the list
        return None
    
    def _get_target_columns(self, columns: Optional[List[str]] = None) -> List[str]:
        """
        Get the list of columns to operate on.
        
        Args:
            columns: List of column names to operate on. If None, uses all xarray columns.
        
        Returns:
            List of column names to operate on
        """
        if columns is None:
            return self._xarray_columns
        
        # Validate that specified columns exist and contain xarray data
        invalid_columns = [col for col in columns if col not in self._xarray_columns]
        if invalid_columns:
            raise ValueError(f"Columns {invalid_columns} are not valid xarray columns. "
                           f"Available xarray columns: {self._xarray_columns}")
        
        return columns
    
    def _should_aggregate_by_geometry(self, dim: Optional[Union[str, List[str]]] = None) -> bool:
        """
        Determine if we should aggregate by geometry (i.e., time dimension was expanded to index).
        
        Args:
            dim: Dimension(s) being reduced over
            
        Returns:
            True if we should group by geometry and aggregate across time rows
        """
        if dim is None:
            return False
        
        dims_to_reduce = [dim] if isinstance(dim, str) else dim
        
        # Check if 'time' is in the dimensions to reduce and if we have a MultiIndex with time
        if 'time' in dims_to_reduce:
            if hasattr(self._obj.index, 'names') and self._obj.index.names:
                # Check if time is one of the index levels
                return 'time' in self._obj.index.names
        
        return False
    
    def _get_geometry_level_name(self) -> Optional[str]:
        """Get the name of the geometry level in the MultiIndex."""
        if hasattr(self._obj.index, 'names') and self._obj.index.names:
            # Look for the level that's not 'time' - this should be the geometry level
            for name in self._obj.index.names:
                if name != 'time':
                    return name
        return None
    
    def _apply_reduction(self, reduction_func: str, dim: Optional[Union[str, List[str]]] = None, 
                        columns: Optional[List[str]] = None, **kwargs):
        """
        Apply a reduction function to specified xarray datasets/dataarrays in the GeoDataFrame.
        
        Args:
            reduction_func: Name of the xarray reduction method (e.g., 'mean', 'sum', 'std')
            dim: Dimension(s) to reduce over. If None, reduces over all dimensions
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments to pass to the reduction function
        
        Returns:
            GeoDataFrame with reduced xarray data
        """
        target_columns = self._get_target_columns(columns)
        
        # Check if we need to aggregate by geometry (time dimension expanded to index)
        if self._should_aggregate_by_geometry(dim):
            return self._apply_temporal_aggregation(reduction_func, dim, target_columns, **kwargs)
        else:
            return self._apply_spatial_reduction(reduction_func, dim, target_columns, **kwargs)
    
    def _apply_temporal_aggregation(self, reduction_func: str, dim: Union[str, List[str]], 
                                  target_columns: List[str], **kwargs):
        """
        Apply aggregation across time by grouping by geometry.
        
        Args:
            reduction_func: Name of the reduction method
            dim: Dimension(s) being reduced (should include 'time')
            target_columns: Columns to operate on
            **kwargs: Additional arguments
        
        Returns:
            GeoDataFrame with time-aggregated data
        """
        geometry_level = self._get_geometry_level_name()
        if geometry_level is None:
            raise ValueError("Could not identify geometry level in MultiIndex")
        
        # Check if specific columns were requested for time aggregation
        if target_columns != self._xarray_columns:
            print("Warning: Cannot aggregate time on a single column. Aggregating all xarray columns instead.")
            target_columns = self._xarray_columns
        
        # Group by geometry level
        grouped = self._obj.groupby(level=geometry_level)
        
        result_data = []
        result_geometries = []
        result_index = []
        
        for geometry_key, group in grouped:
            # For each geometry, collect all xarray objects across time
            # The geometry is the group key itself (from the MultiIndex)
            new_row = {}
            
            for col in target_columns:
                xarray_objects = []
                
                # Collect all xarray objects for this geometry across different times
                for _, row in group.iterrows():
                    xr_data = self._extract_xarray_object(row[col])
                    if xr_data is not None:
                        xarray_objects.append(xr_data)
                
                if xarray_objects:
                    try:
                        # Concatenate along a new 'time' dimension
                        if isinstance(xarray_objects[0], xr.DataArray):
                            # Create time coordinate
                            time_coords = list(range(len(xarray_objects)))
                            concatenated = xr.concat(xarray_objects, dim='time')
                            concatenated = concatenated.assign_coords(time=time_coords)
                        elif isinstance(xarray_objects[0], xr.Dataset):
                            time_coords = list(range(len(xarray_objects)))
                            concatenated = xr.concat(xarray_objects, dim='time')
                            concatenated = concatenated.assign_coords(time=time_coords)
                        else:
                            raise TypeError(f"Unsupported xarray type: {type(xarray_objects[0])}")
                        
                        # Apply the reduction function over the time dimension
                        if hasattr(concatenated, reduction_func):
                            if 'skipna' not in kwargs and reduction_func in ['mean', 'sum', 'std', 'var', 'min', 'max', 'median', 'quantile']:
                                kwargs['skipna'] = True
                            
                            reduced_data = getattr(concatenated, reduction_func)(dim='time', **kwargs)
                            
                            # Check if result should be converted to scalar
                            if isinstance(reduced_data, xr.DataArray) and reduced_data.size == 1:
                                try:
                                    scalar_value = float(reduced_data.values)
                                    reduced_data = scalar_value
                                except (ValueError, TypeError):
                                    pass
                            elif isinstance(reduced_data, xr.Dataset) and len(reduced_data.dims) == 0:
                                try:
                                    vars_list = list(reduced_data.data_vars.keys())
                                    if len(vars_list) == 1:
                                        var_name = vars_list[0]
                                        scalar_value = float(reduced_data[var_name].values)
                                        reduced_data = scalar_value
                                except (ValueError, TypeError, KeyError):
                                    pass
                            
                            # Maintain original format (list vs direct)
                            original_format = group[col].iloc[0]
                            if isinstance(original_format, list):
                                new_row[col] = [reduced_data]
                            else:
                                new_row[col] = reduced_data
                        else:
                            raise AttributeError(f"'{type(concatenated).__name__}' object has no attribute '{reduction_func}'")
                    
                    except Exception as e:
                        print(f"Warning: Could not apply {reduction_func} to geometry {geometry_key}, column {col}: {e}")
                        # Keep the first value as fallback
                        new_row[col] = group[col].iloc[0]
                else:
                    # No xarray data found, keep first value
                    new_row[col] = group[col].iloc[0]
            
            result_data.append(new_row)
            result_geometries.append(geometry_key)
            result_index.append(geometry_key)
        
        # Create result GeoDataFrame
        # Create a normal DataFrame with just the data columns
        result_df = pd.DataFrame(result_data, index=result_index)
        
        # Add geometry as a temporary column
        result_df['_temp_geom'] = result_geometries
        
        # Convert to GeoDataFrame using the temporary geometry column
        result_gdf = gpd.GeoDataFrame(result_df, geometry='_temp_geom')
        
        # Drop the temporary geometry column (the geometry is now properly set as the active geometry)
        result_gdf = result_gdf.drop(columns=['_temp_geom'])
        
        result_gdf.index.name = geometry_level
        
        return result_gdf
    
    def _apply_spatial_reduction(self, reduction_func: str, dim: Optional[Union[str, List[str]]], 
                               target_columns: List[str], **kwargs):
        """
        Apply reduction to spatial dimensions within each xarray object.
        
        Args:
            reduction_func: Name of the reduction method
            dim: Spatial dimension(s) to reduce over
            target_columns: Columns to operate on
            **kwargs: Additional arguments
        
        Returns:
            GeoDataFrame with spatially reduced data
        """
        result_gdf = self._obj.copy()
        
        for col in target_columns:
            new_data = []
            for idx, row in self._obj.iterrows():
                original_value = row[col]
                xr_data = self._extract_xarray_object(original_value)
                
                if xr_data is not None:
                    try:
                        # Apply the reduction function
                        if hasattr(xr_data, reduction_func):
                            # Ensure skipna=True is set by default for most reduction functions
                            if 'skipna' not in kwargs and reduction_func in ['mean', 'sum', 'std', 'var', 'min', 'max', 'median', 'quantile']:
                                kwargs['skipna'] = True
                            reduced_data = getattr(xr_data, reduction_func)(dim=dim, **kwargs)
                            
                            # Check if the result is a scalar and convert to float if so
                            if isinstance(reduced_data, xr.DataArray):
                                if reduced_data.size == 1:
                                    try:
                                        scalar_value = float(reduced_data.values)
                                        reduced_data = scalar_value
                                    except (ValueError, TypeError):
                                        pass
                            elif isinstance(reduced_data, xr.Dataset):
                                try:
                                    if len(reduced_data.dims) == 0:
                                        vars_list = list(reduced_data.data_vars.keys())
                                        if len(vars_list) == 1:
                                            var_name = vars_list[0]
                                            scalar_value = float(reduced_data[var_name].values)
                                            reduced_data = scalar_value
                                except (ValueError, TypeError, KeyError):
                                    pass
                            
                            # Keep the same format as original (list vs direct)
                            if isinstance(original_value, list):
                                new_data.append([reduced_data])
                            else:
                                new_data.append(reduced_data)
                        else:
                            raise AttributeError(f"'{type(xr_data).__name__}' object has no attribute '{reduction_func}'")
                    except Exception as e:
                        # If reduction fails, keep original data
                        print(f"Warning: Could not apply {reduction_func} to row {idx}, column {col}: {e}")
                        new_data.append(original_value)
                else:
                    # If it's not xarray data, keep as is
                    new_data.append(original_value)
            
            result_gdf[col] = new_data
        
        return result_gdf
    
    def mean(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate mean of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
                 If spatial dims like 'x', 'y', reduces within each xarray object.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('mean', dim=dim, columns=columns, **kwargs)
    
    def sum(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate sum of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('sum', dim=dim, columns=columns, **kwargs)
    
    def std(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate standard deviation of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('std', dim=dim, columns=columns, **kwargs)
    
    def var(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate variance of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('var', dim=dim, columns=columns, **kwargs)
    
    def min(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate minimum of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('min', dim=dim, columns=columns, **kwargs)
    
    def max(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate maximum of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('max', dim=dim, columns=columns, **kwargs)
    
    def median(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate median of xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('median', dim=dim, columns=columns, **kwargs)
    
    def quantile(self, q: float, dim: Optional[Union[str, List[str]]] = None, 
                columns: Optional[List[str]] = None, **kwargs):
        """
        Calculate quantile of xarray datasets/dataarrays.
        
        Args:
            q: Quantile to compute (between 0 and 1)
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('quantile', dim=dim, columns=columns, q=q, **kwargs)
    
    def count(self, dim: Optional[Union[str, List[str]]] = None, columns: Optional[List[str]] = None, **kwargs):
        """
        Count non-NaN values in xarray datasets/dataarrays.
        
        Args:
            dim: Dimension(s) to reduce over. If 'time', aggregates across time rows for each geometry.
            columns: List of column names to operate on. If None, operates on all xarray columns
            **kwargs: Additional arguments for the reduction function
        """
        return self._apply_reduction('count', dim=dim, columns=columns, **kwargs)
    
    def to_values(self, columns: Optional[List[str]] = None):
        """
        Extract scalar values from xarray dataarrays and add them as new columns.
        Useful when dataarrays have been reduced to single values.
        
        Args:
            columns: List of column names to operate on. If None, operates on all xarray columns
        
        Returns:
            GeoDataFrame with extracted values as new columns
        """
        result_gdf = self._obj.copy()
        target_columns = self._get_target_columns(columns)
        
        for col in target_columns:
            values_to_add = []
            for idx, row in self._obj.iterrows():
                xr_data = self._extract_xarray_object(row[col])
                if isinstance(xr_data, xr.DataArray):
                    try:
                        if xr_data.size == 1:
                            scalar_value = float(xr_data.values)
                            values_to_add.append(scalar_value)
                        else:
                            values_to_add.append(np.nan)  # Can't convert non-scalar to value
                    except (ValueError, TypeError):
                        values_to_add.append(np.nan)
                else:
                    values_to_add.append(np.nan)
            
            # Add new column with scalar values
            new_col_name = f"{col}_value"
            result_gdf[new_col_name] = values_to_add
        
        return result_gdf
    
    def info(self):
        """Print information about xarray datasets/dataarrays in the GeoDataFrame."""
        print(f"GeoDataFrame shape: {self._obj.shape}")
        print(f"Columns: {list(self._obj.columns)}")
        print(f"Xarray columns: {self._xarray_columns}")
        print(f"Index structure: {self._obj.index.names if hasattr(self._obj.index, 'names') else 'Simple index'}")
        print(f"Geometry column name: {self._obj.geometry.name if hasattr(self._obj.geometry, 'name') else 'No geometry name'}")
        
        if hasattr(self._obj.index, 'names') and 'time' in self._obj.index.names:
            print("Note: Time dimension appears to be expanded into the index.")
            print("Use dim='time' to aggregate across time rows for each geometry.")
        
        for col in self._xarray_columns:
            print(f"\n--- Column: {col} ---")
            sample_data = self._extract_xarray_object(self._obj[col].iloc[0]) if len(self._obj) > 0 else None
            if isinstance(sample_data, xr.Dataset):
                print(f"Type: xarray.Dataset")
                print(f"Variables: {list(sample_data.data_vars.keys())}")
                print(f"Dimensions: {list(sample_data.dims.keys())}")
                print(f"Coordinates: {list(sample_data.coords.keys())}")
            elif isinstance(sample_data, xr.DataArray):
                print(f"Type: xarray.DataArray")
                print(f"Dimensions: {list(sample_data.dims)}")
                print(f"Shape: {sample_data.shape}")
                print(f"Data type: {sample_data.dtype}")


# terrakio_core/__init__.py
"""
Terrakio Core

Core components for Terrakio API clients.
"""

__version__ = "0.4.6"

from .async_client import AsyncClient
from .sync_client import SyncClient as Client
from . import accessors


__all__ = [
    "AsyncClient", 
    "Client"
]


import geopandas as gpd
import shapely.geometry
import json

def escape_newline(string):
    if isinstance(string, list):
        return [s.replace('\\n', '\n') for s in string]
    else:
        return string.replace('\\n', '\n')

def get_bounds(aoi, crs, to_crs = None):
    aoi : gpd.GeoDataFrame = gpd.read_file(aoi)
    aoi = aoi.set_crs(crs, allow_override=True)
    if to_crs:
        aoi = aoi.to_crs(to_crs)
    bounds = aoi.geometry[0].bounds
    return *bounds, aoi

def tile_generator(x_min, y_min, x_max, y_max, aoi, crs, res, tile_size, expression, output, fully_cover=True):
    i_max = int((x_max-x_min)/(tile_size*res))
    j_max = int((y_max-y_min)/(tile_size*res))
    if fully_cover:
        i_max += 1
        j_max += 1
    for j in range(0, int(j_max)):
        for i in range(0, int(i_max)):
            x = x_min + i*(tile_size*res)
            y = y_max - j*(tile_size*res)
            bbox = shapely.geometry.box(x, y-(tile_size*res), x + (tile_size*res), y)
            if not aoi.geometry[0].intersects(bbox):
                continue
            feat  = {"type": "Feature", "geometry": bbox.__geo_interface__}
            data = {
                "feature": feat,
                "in_crs": crs,
                "out_crs": crs,
                "resolution": res,
                "expr" : expression,
                "output" : output,
            }
            yield data, i , j


def tiles(
    name: str,
    aoi : str, 
    expression: str = "red=S2v2#(year,median).red@(year =2024) \n red",
    output: str = "netcdf",
    tile_size : float = 512,
    crs : str = "epsg:3577",
    res: float = 10,
    region : str = "eu",
    to_crs: str = None,
    fully_cover: bool = True,
    overwrite: bool = False,
    skip_existing: bool = False,
    non_interactive: bool = False,
):
    
    reqs = []
    x_min, y_min, x_max, y_max, aoi = get_bounds(aoi, crs, to_crs)

    if to_crs is None:
        to_crs = crs
    for tile_req, i, j in tile_generator(x_min, y_min, x_max, y_max, aoi, to_crs, res, tile_size, expression, output, fully_cover):
        req_name = f"{name}_{i:02d}_{j:02d}"
        reqs.append({"group": "tiles", "file": req_name, "request": tile_req})

    count = len(reqs)
    groups = list(set(dic["group"] for dic in reqs))

    body = {
        "name" : name,
        "output" : output,
        "region" : region,
        "size" : count,
        "overwrite" : overwrite,
        "non_interactive": non_interactive,
        "skip_existing" : skip_existing,
    }
    request_json = json.dumps(reqs)
    manifest_json = json.dumps(groups)

    return body, request_json, manifest_json


from typing import Dict, Any, Optional
from ..helper.decorators import require_token, require_api_key, require_auth
class SpaceManagement:
    def __init__(self, client):
        self._client = client

    @require_api_key
    def get_total_space_used(self) -> Dict[str, Any]:
        """
        Get total space used by the user.
        
        Returns:
            Dict[str, Any]: Total space used by the user.

        Raises:
            APIError: If the API request fails
        """
        return self._client._terrakio_request("GET", "/users/jobs")

    @require_api_key
    def get_space_used_by_job(self, name: str, region: str) -> Dict[str, Any]:
        """
        Get space used by a specific job.

        Args:
            name: The name of the job
            region: The region of the job

        Returns:
            Dict[str, Any]: Space used by the job.
        
        Raises:
            APIError: If the API request fails
        """
        params = {"region": region}
        return self._client._terrakio_request("GET", f"/users/jobs/{name}", params=params)
    
    @require_api_key
    def delete_user_job(self, name: str, region: str) -> Dict[str, Any]:
        """
        Delete a user job by name and region.
        
        Args:
            name: The name of the job
            region: The region of the job
            
        Returns:
            Dict[str, Any]: Response from the delete operation.
        
        Raises:
            APIError: If the API request fails
        """
        params = {"region": region}
        return self._client._terrakio_request("DELETE", f"/users/jobs/{name}", params=params)

    @require_api_key
    def delete_data_in_path(self, path: str, region: str) -> Dict[str, Any]:
        """
        Delete data in a GCS path for a given region.
        
        Args:
            path: The GCS path to delete data from
            region: The region where the data is located
            
        Returns:
            Dict[str, Any]: Response from the delete operation.
        
        Raises:
            APIError: If the API request fails
        """
        params = {"path": path, "region": region}
        return self._client._terrakio_request("DELETE", "/users/jobs", params=params)